//
//   ComPort.h
//

#ifndef ComPortDef
#define ComPortDef 1

#include <windows.h>
#include <stdio.h>
#include <vector>

#include "..\UtilsLib\main_classes.h"
#include "..\UtilsLib\status_bar.h"
#include "..\UtilsLib\controls.h"
#include "..\UtilsLib\lng.h"
#include "..\UtilsLib\abstract_model.h"
#include "..\UtilsLib\ByteStream.h"

#include "config.h"


#define  Terminal_ComPort_Adapter_CLASS_NAME    "Terminal_ComPort_Adapter"
#define  Terminal_ComPort_Adapter_OBJECT_TYPE   "Terminal_ComPort_Adapter"


class Terminal_ComPort_Adapter : protected ToolClass, protected TicksUserClass {
friend dword WINAPI ComPortWriteThread( LPVOID pArg );
friend dword WINAPI ComPortReadThread(  LPVOID pArg );
/*
friend void ComPortWriteThread( LPVOID pArg );
friend void ComPortReadThread(  LPVOID pArg );
//*/
public:
	static const char* const CLASS_NAME; 
	
	static const char* const OBJECT_TYPE; 
	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;


protected:
	virtual ~Terminal_ComPort_Adapter();

//	IntegerConstant  nC_SecondsToWaitAtPrompt;
//	StringParameter  sP_ScriptFile;
//	StateVarType     St_Load;

	void ByteReceiver( PVOID pData, SocketClass *pSender );

	Input_SocketClass         ByteInputSocket;
	Output_SocketClass        ByteOutputSocket;
	
	void OutputSocket_ReportBaudRate_CommandProcessor( PVOID, SocketClass *pSender )
	{
		ByteOutputSocket.SendReply( acReply_BaudRateReported, (PVOID)nC_BaudRate.Get(), pSender );
	}

	void Initializer();

	void Tick();

	dword WINAPI AdapterWriteThread();
	HANDLE hWriteThread;
	dword WINAPI AdapterReadThread();
	HANDLE hReadThread;

	safe_string  sComPort_Output;
	std::string  sWriteThread_Output;
	
	int nMesCount;

private:
	Terminal_ComPort_Adapter( ConfigInterface *pConfigI, const char *acAlias );

	static const NamedValType                       aBooleanVals[];
	static const std::vector< const NamedValType >  vBooleanVals;
	
	HANDLE hComPort;
	
	StringConstant  sC_PortName;
	
	bool  bCOM_OutputBreak;
	bool  bCOM_OutputBreakOff;
	void  BreakOn_CommandProcessor( PVOID, SocketClass* );	
	void  BreakOff_CommandProcessor( PVOID, SocketClass* );
	void  ClearFIFO_CommandProcessor( PVOID, SocketClass* );	


	static const NamedValType                       aValidBaudRates[];
	static const std::vector< const NamedValType >  vValidBaudRates;
	IntegerConstant                                 nC_BaudRate;

	static const NamedValType                       aValidDtrControls[];
	static const std::vector< const NamedValType >  vValidDtrControls;
	IntegerConstant                                 nC_fDtrControl;

	static const NamedValType                       aValidRtsControls[];
	static const std::vector< const NamedValType >  vValidRtsControls;
	IntegerConstant                                 nC_fRtsControl;

	static const NamedValType                       aValidParitys[];
	static const std::vector< const NamedValType >  vValidParitys;
	IntegerConstant                                 nC_Parity;
	
	static const NamedValType                       aValidStopBits[];
	static const std::vector< const NamedValType >  vValidStopBits;
	IntegerConstant                                 nC_StopBits;

	IntegerConstant                                 nC_MinimalBreakTimeMKS;
	
	StateVarType     St_StopReading;
	
	StateVarType     St_ShowPortUse;
	void             ShowPortUse_StateCallback( const char*, int nState );
	bool             bShowCPS;
	
	IntegerConstant  nC_ByteSize;

	IntegerConstant  nC_fParity;
	IntegerConstant  nC_fOutxCtsFlow;
	IntegerConstant  nC_fOutxDsrFlow;
	IntegerConstant  nC_fDsrSensitivity;
	IntegerConstant  nC_fTXContinueOnXoff;
	IntegerConstant  nC_fOutX;
	IntegerConstant  nC_fInX;
	IntegerConstant  nC_fErrorChar;
	IntegerConstant  nC_fNull;
	IntegerConstant  nC_fAbortOnError;
	
	IntegerConstant  nC_XonLim;
	IntegerConstant  nC_XoffLim;

	IntegerConstant  nC_XonChar;
	IntegerConstant  nC_XoffChar;
	IntegerConstant  nC_ErrorChar;
	
	dword  dwBytesSent;
	dword  dwPrevBytesSent;
	dword  dwBytesRecived;
	dword  dwPrevBytesRecived;
	dword  dwLastPrintTC;
	
	WorkfieldOwnerStatusBarClass  *pStatusBar;
	
	std::vector<HANDLE>  vhWriteThreadEvents;
	std::vector<HANDLE>  vhReadThreadEvents;
	
	HANDLE hEvAbort;
	HANDLE hEvOutput;
	
	LARGE_INTEGER  OutBreakOn_Cnt;
	dword          dwTargetOutBreakTimeMKS;

	static void OutBreakOff_ThreadStarter( PVOID pObject )
	{
		return ((Terminal_ComPort_Adapter*)pObject)->OutBreakOff_Thread();
	}
	void OutBreakOff_Thread();
	

	bool           bCOM_InputBreakOff;
	LARGE_INTEGER  InBreakOn_Cnt;
	void           BreakOff_Delay();
	

	static unsigned int nObjectsCount;
};

#endif
