//
//  Main_module.h  -  Header file for Main_module.
//
#ifndef Main_moduleDefs
#define Main_moduleDefs 1

#include <windows.h>
#include <string>
#include <vector>
#include <deque>
#include <algorithm>
#include <functional>

#include "config.h"
#include "SBarPart_Objects.h"
#include "ODT_Loader.h"
#include "IP.h"
#include "ComPort.h"
#include "UKNCcomSender.h"
#include "UKNC_NET_Loader.h"
#include "LogSpeedTester.h"

#include "..\UtilsLib\FastDelegate.h"
#include "..\UtilsLib\main_classes.h"
#include "..\UtilsLib\abstract_model.h"
#include "..\UtilsLib\PDP11A16_classes.h"
#include "..\UtilsLib\console_class.h"
#include "..\UtilsLib\workfield_class.h"
#include "..\UtilsLib\WFTabs.h"
#include "..\UtilsLib\FastDelegate.h"


///

#define  PDP11A16_BusClass_CLASS_NAME    "PDP11A16_Bus"
#define  PDP11A16_BusClass_OBJECT_TYPE   PDP11A16


class PDP11A16_BusClass : protected AbstractBusClass, protected PDP11A16_BusInterface {
public:
	static const char* const CLASS_NAME; // = "PDP11A16_Bus";

	static const char* const OBJECT_TYPE;     // "PDP11A16"
	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;

protected:
	virtual ~PDP11A16_BusClass();

	virtual	ObjectClass*   Get_pOwner()	{ return (ObjectClass*)this; }

	PDP11A16_CPU_BASE_Interface  *p16CPU_I;

	StateVarType  St_HALT;
	StateVarType  St_Timer;

	void  Timer_StateCallback( const char*, int nState );
	void  HALT_StateCallback(  const char*, int nState );

	byte                              *aMemory;
	ObjectClass*                      *IO_map;
	IO_FunDelegateType                *IO_map2;
	bool                              *abRAM;
	bool                              *abReadAccess;
	bool                              *abWriteAccess;
	int                               *anAccessLatencyNS;
	int                               *anAccessLatencyClocks;
	
	virtual bool PrepareEmulation();

	virtual void Get_MemoryAccessData( MemoryAccessDataType *pMemData )
	{
		pMemData->paMemory = aMemory;
		pMemData->pIO_map  = IO_map;
		pMemData->pIO_map2 = IO_map2;
		pMemData->pabRAM   = abRAM;
		pMemData->pabReadAccess  = abReadAccess;
		pMemData->pabWriteAccess = abWriteAccess;
		pMemData->panAccessLatencyNS = anAccessLatencyNS;
		pMemData->panAccessLatencyClocks = anAccessLatencyClocks;
	}
	virtual bool Is_V100_DeviceInstalled(){ return b_V100_DeviceInstalled; }
	virtual void Init()
	{
		for( int i = 0 ; i < vpInstalledDevices.size() ; i++ )
		{
			if( vpInstalledDevices[i] == p16CPU_I->Get_pOwner() ) { continue; }
			PDP11A16_DeviceInterface *pDevI = (PDP11A16_DeviceInterface*)vpInstalledDevices[i]->Get_pInterface( PDP11A16_DeviceInterface::CLASS_NAME );
			pDevI->Init();
		}
	}
	
	virtual bool SetIRQ( word uVector, word uBR_Line, word uBR_LinePosition, bool bClearedByInit = true, PDP11A16_DeviceInterface* pDevI = NULL )
	{
		if( !pCPU ) { return false; }
		if( uVector != 0100 || !St_Pause() )
		{
			return p16CPU_I->SetIRQ( uVector, uBR_Line, uBR_LinePosition, bClearedByInit, pDevI );
		}
		return false;
	}
	virtual bool ClearIRQ( word uVector )
	{
		if( !pCPU ) { return false; }
		return p16CPU_I->ClearIRQ( uVector );
	}
	virtual void SetFastCallback( fastdelegate::FastDelegate0<> Callback, int nSteps )
	{
		if( !pCPU ) { return; }
		return p16CPU_I->SetFastCallback( Callback, nSteps );
	}
	
	int  InstallDeviceCheck( ObjectClass *pObject );
	void InstallDevice( ObjectClass *pObject );
	
	void Power_StateCallback( const char*, int nState );

   	bool   b_V100_DeviceInstalled;

private:
	PDP11A16_BusClass( ConfigInterface *pConfigI, const char *acAlias );
	
	int  LinkCheck( SocketClass*, SocketClass* );
	void LinkDone( SocketClass*, SocketClass* );
	
	static unsigned int nObjectsCount;
};

//////////////////

#define  CON_Class_CLASS_NAME    "CON"
#define  CON_Class_OBJECT_TYPE   "Console"


class CON_Class : protected ToolClass, protected WorkfieldOwnerClass {
public:
	static const char* const CLASS_NAME; // "CON"
	
	static const char* const OBJECT_TYPE; // "Console"
	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;

	virtual void WF_Tick();
	virtual LRESULT CALLBACK Parse_WM_MOUSEWHEEL( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );

protected:
	virtual ~CON_Class(){}

	DWindowInitData InitData;
	
	Console      *pCon;
	StateVarType  St_Log;
	void          Log_StateCallback( const char*, int nState )
	{
		if( nState < 0 ) { return; }

		if( nState )
		{
			if(pCon) { pCon->StartLog(); }
		}
		else
		{
			if(pCon) { pCon->StopLog(); }
		}
	}
	virtual void  Tick() { if(pCon) { pCon->FlushLogData(); } }
	
	Input_SocketClass   TextInputSocket;
	Input_SocketClass   ByteInputSocket;
	Output_SocketClass  ByteOutputSocket;
	
	IntegerConstant nC_DumpMode;
	IntegerConstant nC_DumpCols;
	IntegerConstant nC_DumpRadix;
	
	static const int                       anValidDumpRadix[];
	static const std::vector< const int >  vnValidDumpRadix; 
	
	StringConstant  sC_LeftDumpSeparator;
	StringConstant  sC_RightDumpSeparator;
	
	byte  cLeftDumpSeparator;
	byte  cRightDumpSeparator;

	virtual void   ThemeChanged(){}
	virtual void   ResizeToRECT( RECT rc )
	{
		if(pCon) { pCon->Resize( rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top ); }
	}
	virtual HWND   GetFocusHWND(){ return NULL; }

	virtual LRESULT CALLBACK Parse_Keyboard_WMs( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
	{
		if( msg == WM_CHAR ) 
		{
			byte ch = (byte)wParam;
			ByteOutputSocket.Send( (PVOID)ch );
		}
		return 0;
	}
//
private:
	CON_Class( ConfigInterface *pConfigI, const char *acAlias );
	void TextReceiver( PVOID pData, SocketClass *pSender )
	{
		if( pCon ) { pCon->StrOut( (char*)pData ); }
	}
	void BadFrame_CommandProcessor( PVOID ch, SocketClass *pSender )
	{
		return ByteReceiver( PVOID(((int)ch)|01000), pSender );
	}
	void ByteReceiver( PVOID ch, SocketClass *pSender );
	
	int nDumpRowLength;
	std::string sDumpRow;
	
	static unsigned int nObjectsCount;
};


//////////////////

#define  BooterClass_CLASS_NAME    "Booter"
#define  BooterClass_OBJECT_TYPE   BooterClass_CLASS_NAME

class BooterClass : protected PDP11A16_DeviceClass {
public:
	static const char* const CLASS_NAME; //  "Booter"
	
	static const char* const OBJECT_TYPE; // "Booter"
	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;

protected:
	virtual ~BooterClass(){}
	
	StringConstant   sC_BootDeviceAlias;
	IntegerConstant  nC_BootDeviceUnit;
	StringConstant   sC_WarmBootSupportedSystem_1;
	IntegerConstant  nC_WarmBootSystem;
	IntegerConstant  nC_SetDateOnly;
	IntegerConstant  nC_RT11_LineClockHz;
	
	StateVarType     St_Boot;
	
private:
	BooterClass( ConfigInterface *pConfigI, const char *acAlias );
	void   Initializer();
	void   Boot();

	void Boot_StateCallback( const char*, int );
	
	void PrepareEmulation() {}

	static unsigned int nObjectsCount;
};



#endif
