//
//   SBarPart_Objects.h
//

#ifndef SBarPart_ObjectsDef
#define SBarPart_ObjectsDef 1

#include <windows.h>
#include <stdio.h>
#include <vector>

#include "..\UtilsLib\main_classes.h"
#include "..\UtilsLib\status_bar.h"
#include "..\UtilsLib\controls.h"
#include "..\UtilsLib\lng.h"
#include "..\UtilsLib\abstract_model.h"

#include "config.h"



#define  SBarPart_ObjectInterface_CLASS_NAME  "SBarPart_ObjectInterface"
#define  acStatusBarPart_SuperType            "StatusBarPart"

const dword  SBarPart_ObjectInterface_Generation = 1;
const dword  SBarPart_ObjectInterface_Version    = 1;


class SBarPart_ObjectInterface : public InterfaceClass {
public:
	static const char* const CLASS_NAME; // "SBarPart_ObjectInterface"
	
	virtual bool Place( WorkfieldOwnerStatusBarClass *pStatusBar )
	{
		bool bOK;
		for( int nSize = vdPlaceControllers.size() , i = 0 ; i < nSize ; i++ )
		{
			bOK = vdPlaceControllers[i]( pStatusBar );
			if( !bOK ) { break; }
		}
		return bOK;
	}
//
protected:
	SBarPart_ObjectInterface( ObjectClass *pOwner ):InterfaceClass( pOwner, CLASS_NAME, SBarPart_ObjectInterface_Generation, SBarPart_ObjectInterface_Version ), pStatusBar(NULL){}
	virtual ~SBarPart_ObjectInterface(){}

	WorkfieldOwnerStatusBarClass *pStatusBar;

	std::vector< fastdelegate::FastDelegate1<WorkfieldOwnerStatusBarClass*,bool> > vdPlaceControllers;
};


class SBarPart_ObjectClass : virtual protected ObjectClass, public SBarPart_ObjectInterface {
protected:
	virtual ~SBarPart_ObjectClass(){}
	
	SBarPart_ObjectClass( ConfigInterface *pConfig_I ):
	ObjectClass( "SBarPart_ObjectClass" ),
	SBarPart_ObjectInterface( (ObjectClass*)this ),
	nPosition(1),
	nPriority(0),
	nHidden(0)
	{
		vdPlaceControllers.push_back( fastdelegate::MakeDelegate(this, &SBarPart_ObjectClass::SBarPart_PlaceController) );
	
		char acBuf[2048];
		char acDefVal[128];
		sprintf(acDefVal,"nPosition[%i] | nPriority[%i] | bHidden[%i]", nPosition, nPriority, nHidden );
		pConfigI->ReadIniStr( this, "SB_Part_Data", acBuf, acDefVal );
		
		sscanf( acBuf, "nPosition[%d] %*c nPriority[%d] %*c bHidden[%d]", &nPosition, &nPriority, &nHidden );
	}
	
	int  nPosition;
	int  nPriority;
	int  nHidden;

	std::string sPartName;

private:

	bool SBarPart_PlaceController( WorkfieldOwnerStatusBarClass *pStatusBAR )
	{
		if( pStatusBar )
		{
			char str[512];
			sprintf(str, LNG("StatusBarPart [%s] is already attached to [%s] StatusBar !!!"), sAlias.c_str(), pStatusBar->Get_pOwner()->Get_sAlias().c_str() );
			sLastError = str;
			return false;
		}
		
		if( WorkfieldOwnerStatusBarClass_Generation != pStatusBAR->Get_dwGeneration() )
		{
			sLastError = acIncompatible_Generations_of_class( "WorkfieldOwnerStatusBarClass" );
			return false;
		}
		
		pStatusBar = pStatusBAR;
		return true;
	}
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////

#define  SB_StatePushButton_CLASS_NAME        "SB_StatePushButton"
#define  SB_StatePushButton_OBJECT_TYPE       SB_StatePushButton_CLASS_NAME


class SB_StatePushButton : protected SBarPart_ObjectClass {
public:
	static const char* const CLASS_NAME;  // "SB_StatePushButton"
	static const char* const OBJECT_TYPE; // "SB_StatePushButton"

	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;
	

protected:
	virtual ~SB_StatePushButton()
	{
		if( pB_State ){ pB_State->Destroy(); }
	}
	
	
private:
	SB_StatePushButton( ConfigInterface *pConfig_I, const char *acAlias );

	bool PlaceController( WorkfieldOwnerStatusBarClass *pStatusBar );
	bool Object_AttachController( ObjectClass *pObject );

	void SetLanguage( const char* )
	{
		if( pB_State ) 
		{
			pB_State->SetON_Text(  LNG( sButtonON_Text.c_str()  ) );
			pB_State->SetOFF_Text( LNG( sButtonOFF_Text.c_str() ) );
		}
		if( pStatusBar && sTextFieldText.size() )
		{
			pStatusBar->StrOut( LNG( sTextFieldText.c_str() ), nPartHandle );
		}
	}
	//
	int                   nTextFieldWidth;
	std::string           sTextFieldText;
	StatePushButtonClass *pB_State;

	int          nButtonPushMode;
	std::string  sButtonON_Text;
	std::string  sButtonOFF_Text;
	int          nButtonWidth;
	int          nButtonHeight;
	int          nButtonFontSize;
	int          nButtonFontWeight;
	bool         bInverted;
	

	std::string                  sStateName;
	std::vector <StateVarType*>  vpStateVars;
	int                          nPartHandle;
	
	static unsigned int nObjectsCount;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////

#define  SB_StateFlashButton_CLASS_NAME      "SB_StateFlashButton"
#define  SB_StateFlashButton_OBJECT_TYPE     SB_StateFlashButton_CLASS_NAME


class SB_StateFlashButton : protected SBarPart_ObjectClass {
public:
	static const char* const CLASS_NAME;  // "SB_StateFlashButton"
	static const char* const OBJECT_TYPE; // "SB_StateFlashButton"

	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;
	

protected:
	virtual ~SB_StateFlashButton()
	{
		if( pB_State ){ pB_State->Destroy(); }
	}
	
private:
	SB_StateFlashButton( ConfigInterface *pConfig_I, const char *acAlias );

	bool PlaceController( WorkfieldOwnerStatusBarClass *pStatusBar );
	bool Object_AttachController( ObjectClass *pObject );

	void SetLanguage( const char* )
	{
		if( pB_State ) 
		{
			pB_State->SetON_Text(  LNG( sButtonON_Text.c_str()  ) );
			pB_State->SetOFF_Text( LNG( sButtonOFF_Text.c_str() ) );
		}
		if( pStatusBar && sTextFieldText.size() )
		{
			pStatusBar->StrOut( LNG( sTextFieldText.c_str() ), nPartHandle );
		}
	}
	//
	void LightsOFF_Controller( const char*, int nState )
	{
		if( nState == NOT_EXIST || !pB_State ) { return; }

		if( nState == OFF || nState == DISABLED_OFF )
		{
			pB_State->DisableOnColors();
		}
		else
		{
			pB_State->EnableOnColors();
		}
	}
	//
	int                   nTextFieldWidth;
	std::string           sTextFieldText;
	StateFlashButtonClass *pB_State;
	
	int          nButtonPushMode;
	std::string  sButtonON_Text;
	std::string  sButtonOFF_Text;
	int          nButtonWidth;
	int          nButtonHeight;
	int          nButtonFontSize;
	int          nButtonFontWeight;
	bool         bInverted;
	
	int            nPushed_Text_X_Offset;
	int            nPushed_Text_Y_Offset;
	int            nPopped_Text_X_Offset;
	int            nPopped_Text_Y_Offset;
	int            nUserText_X_Offset;
	COLORREF       dwON_TextColor;
	COLORREF       dwOFF_TextColor;
	COLORREF       dwON_Background;
	COLORREF       dwOFF_Background;
	bool           bButtonLike_PoppedBorder;
	bool           bButtonLike_PushedBorder;
	bool           bTextXCenter;	


	std::string    sStateName;
	StateVarType  *pStateVar;

	std::string    sLightsOFF_StateName;
	StateVarType  *pLightsOFF_StateVar;

	int            nPartHandle;

	static unsigned int nObjectsCount;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////

#define  SBarPartControl_ObjectInterface_CLASS_NAME   "SBarPartControl_ObjectInterface"

const dword  SBarPartControl_ObjectInterface_Generation = 1;
const dword  SBarPartControl_ObjectInterface_Version    = 1;


class SBarPartControl_ObjectInterface : public InterfaceClass {
public:
	static const char* const CLASS_NAME; // "SBarPartControl_ObjectInterface"

	virtual ControlsClass*  Get_pControl() const { return pControl; }
	virtual void            Set_pStatusBar( WorkfieldOwnerStatusBarClass *pStatusBAR )=0;

protected:
	SBarPartControl_ObjectInterface( ObjectClass *pOwner ):
	InterfaceClass( pOwner, CLASS_NAME, SBarPartControl_ObjectInterface_Generation, SBarPartControl_ObjectInterface_Version ), 
	pControl(NULL), 
	pStatusBar(NULL)
	{}
	virtual ~SBarPartControl_ObjectInterface(){}

	ControlsClass *pControl;
	WorkfieldOwnerStatusBarClass *pStatusBar;
};


class SBarPartControl_ObjectClass : protected ObjectClass, protected SBarPartControl_ObjectInterface {
protected:
	virtual ~SBarPartControl_ObjectClass(){}
	
	SBarPartControl_ObjectClass( ConfigInterface *pConfigI, const char *acSuperType, const char *acType, const char *acAlias, InternalsType *pInternalsRoot ):
	ObjectClass( pConfigI, acSuperType, acType, acAlias, pInternalsRoot ),
	SBarPartControl_ObjectInterface( (ObjectClass*)this )
	{
	}
	
	std::string sControlName;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////

#define  acStatusBarPartControl_SuperType   "StatusBarPartControl"

#define  SB_Part_CLASS_NAME    "SB_Part"
#define  SB_Part_OBJECT_TYPE   SB_Part_CLASS_NAME

class SB_Part : protected SBarPart_ObjectClass {
public:
	static const char* const CLASS_NAME;  // "SB_Part"
	static const char* const OBJECT_TYPE; // "SB_Part"

	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;
	

protected:
	virtual ~SB_Part(){}
	
private:
	SB_Part( ConfigInterface *pConfig_I, const char *acAlias );

	bool PlaceController( WorkfieldOwnerStatusBarClass *pStatusBar );
	bool SBarPartControl_AttachController( ObjectClass *pObject );
	
	void SetLanguage( const char* )
	{
		if( !pStatusBar ) { return; }
		
		if( sTextFieldText.size() )
		{
			pStatusBar->StrOut( LNG( sTextFieldText.c_str() ), nPartHandle );
		}
	}
	
	int                   nTextFieldWidth;
	std::string           sTextFieldText;

	std::vector< SBarPartControl_ObjectInterface* >  vpControls;

	int            nPartHandle;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////

#define  SBPC_StatePushButton_CLASS_NAME    "SBPC_StatePushButton"
#define  SBPC_StatePushButton_OBJECT_TYPE   SBPC_StatePushButton_CLASS_NAME

class SBPC_StatePushButton : protected SBarPartControl_ObjectClass {
public:
	static const char* const CLASS_NAME;  // "SBPC_StatePushButton"
	static const char* const OBJECT_TYPE; // "SBPC_StatePushButton"

	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;
	

protected:
	virtual ~SBPC_StatePushButton()
	{
		if( pB_State ){ pB_State->Destroy(); }
	}
	
private:
	SBPC_StatePushButton( ConfigInterface *pConfig_I, const char *acAlias );

	bool Object_AttachController( ObjectClass *pObject );
	void Set_pStatusBar( WorkfieldOwnerStatusBarClass *pStatusBAR );
	
	void SetLanguage( const char* )
	{
		if( pB_State ) 
		{
			pB_State->SetON_Text(  LNG( sButtonON_Text.c_str()  ) );
			pB_State->SetOFF_Text( LNG( sButtonOFF_Text.c_str() ) );
		}
	}	
	//
	
	StatePushButtonInitData  InitData;
	StatePushButtonClass    *pB_State;

	int          nButtonPushMode;
	std::string  sButtonON_Text;
	std::string  sButtonOFF_Text;
	int          nButtonWidth;
	int          nButtonHeight;
	int          nButtonFontSize;
	int          nButtonFontWeight;
	bool         bInverted;
	
	std::string    sStateName;
	StateVarType  *pStateVar;
	
	static unsigned int nObjectsCount;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////

#define  SBPC_StateFlashButton_CLASS_NAME     "SBPC_StateFlashButton"
#define  SBPC_StateFlashButton_OBJECT_TYPE    SBPC_StateFlashButton_CLASS_NAME

class SBPC_StateFlashButton : protected SBarPartControl_ObjectClass {
public:
	static const char* const CLASS_NAME;  // "SBPC_StateFlashButton"
	static const char* const OBJECT_TYPE; // "SBPC_StateFlashButton"

	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;
	

protected:
	virtual ~SBPC_StateFlashButton()
	{
		if( pB_State ){ pB_State->Destroy(); }
	}
	
private:
	SBPC_StateFlashButton( ConfigInterface *pConfig_I, const char *acAlias );

	bool Object_AttachController( ObjectClass *pObject );
	void Set_pStatusBar( WorkfieldOwnerStatusBarClass *pStatusBAR );
	
	void SetLanguage( const char* )
	{
		if( pB_State ) 
		{
			pB_State->SetON_Text(  LNG( sButtonON_Text.c_str()  ) );
			pB_State->SetOFF_Text( LNG( sButtonOFF_Text.c_str() ) );
		}
	}	
	//
	void LightsOFF_Controller( const char*, int nState )
	{
		if( nState == NOT_EXIST || !pB_State ) { return; }

		if( nState == OFF || nState == DISABLED_OFF )
		{
			pB_State->DisableOnColors();
		}
		else
		{
			pB_State->EnableOnColors();
		}
	}
	//
	
	StateFlashButtonInitData  InitData;
	StateFlashButtonClass    *pB_State;

	int          nButtonPushMode;
	std::string  sButtonON_Text;
	std::string  sButtonOFF_Text;
	int          nButtonWidth;
	int          nButtonHeight;
	int          nButtonFontSize;
	int          nButtonFontWeight;
	bool         bInverted;
	
	int            nPushed_Text_X_Offset;
	int            nPushed_Text_Y_Offset;
	int            nPopped_Text_X_Offset;
	int            nPopped_Text_Y_Offset;
	int            nUserText_X_Offset;
	COLORREF       dwON_TextColor;
	COLORREF       dwOFF_TextColor;
	COLORREF       dwON_Background;
	COLORREF       dwOFF_Background;
	bool           bButtonLike_PoppedBorder;
	bool           bButtonLike_PushedBorder;
	bool           bTextXCenter;
	
	std::string    sStateName;
	StateVarType  *pStateVar;

	std::string    sLightsOFF_StateName;
	StateVarType  *pLightsOFF_StateVar;
	
	static unsigned int nObjectsCount;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////

#define  SB_CPU_MIPS_Limit_CLASS_NAME     "SB_CPU_MIPS_Limit"
#define  SB_CPU_MIPS_Limit_OBJECT_TYPE    SB_CPU_MIPS_Limit_CLASS_NAME

class SB_CPU_MIPS_Limit : protected SBarPart_ObjectClass {
public:
	static const char* const CLASS_NAME;  // "SB_CPU_MIPS_Limit"
	static const char* const OBJECT_TYPE; // "SB_CPU_MIPS_Limit"

	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;
	

protected:
	virtual ~SB_CPU_MIPS_Limit()
	{
		if( pCPU_SpeedMenu ){ pCPU_SpeedMenu->Destroy(); }
	}
	
private:
	SB_CPU_MIPS_Limit( ConfigInterface *pConfig_I, const char *acAlias );

	bool PlaceController( WorkfieldOwnerStatusBarClass *pStatusBar );
	bool CPU_AttachController( ObjectClass *pObject );
	
	void SetLanguage( const char* )
	{
		if( pStatusBar && sTextFieldText.size() )
		{
			pStatusBar->StrOut( LNG( sTextFieldText.c_str() ), nPartHandle );
		}
	}
	//

	int          nTextFieldWidth;
	std::string  sTextFieldText;
	
	IntegerParameterPushMenuInitData  InitData;
	IntegerParameterPushMenuClass    *pCPU_SpeedMenu;

	ObjectClass       *pCPU;
	IntegerParameter  *pNP_CPU_SpeedLimit_KIPS;

	
	void Make_CPU_SpeedMenu();
	
	int  nPartHandle;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////

#define   SB_CPU_MIPS_CLASS_NAME    "SB_CPU_MIPS"
#define   SB_CPU_MIPS_OBJECT_TYPE   SB_CPU_MIPS_CLASS_NAME

class SB_CPU_MIPS : protected SBarPart_ObjectClass, protected TicksUserClass {
public:
	static const char* const CLASS_NAME;  // "SB_CPU_MIPS"
	static const char* const OBJECT_TYPE; // "SB_CPU_MIPS"

	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;
	

protected:
	virtual ~SB_CPU_MIPS();
	
private:
	SB_CPU_MIPS( ConfigInterface *pConfig_I, const char *acAlias );

	bool PlaceController( WorkfieldOwnerStatusBarClass *pStatusBar );
	bool CPU_AttachController( ObjectClass *pObject );
	
	void SetLanguage( const char* )
	{
		if( pStatusBar && sTextFieldText.size() )
		{
			pStatusBar->StrOut( LNG( sTextFieldText.c_str() ), nPartHandle );
		}
	}
	//
	void  Tick();
	
	dword uLastTickCount;
	dword uLastShowValTC;
	dword uLoops;
	qword qwLastStepCount;

	dword uLastClocksTC;
	qword qwLastClocks;
	float ff;
	
	FlashButtonInitData   fbInitData;
	FlashButtonClass     *pFB_Text;

	int                   nTextFieldWidth;
	std::string           sTextFieldText;
	std::string           sF_text;
	
	bool  bShow_Fclk;
	
	ObjectClass            *pCPU;
	AbstractCPU_Interface  *pCPU_I;
	
	int   nPartHandle;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////

#define  SB_StorageMediaButton_CLASS_NAME     "SB_StorageMediaButton"
#define  SB_StorageMediaButton_OBJECT_TYPE    SB_StorageMediaButton_CLASS_NAME

class SB_StorageMediaButton : protected SBarPart_ObjectClass, protected TicksUserClass {
public:
	static const char* const CLASS_NAME;  // "SB_StorageMediaButton"
	static const char* const OBJECT_TYPE; // "SB_StorageMediaButton"

	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;
	

protected:
	virtual ~SB_StorageMediaButton()
	{
		if( pUnitsMenu )   { pUnitsMenu->Destroy();   }
		if( pMediaButton ) { pMediaButton->Destroy(); }
		if( pDevicesMenu ) { pDevicesMenu->Destroy(); }
	}
	
private:
	SB_StorageMediaButton( ConfigInterface *pConfig_I, const char *acAlias );

	bool PlaceController( WorkfieldOwnerStatusBarClass *pStatusBar );
	bool StorageMediaUser_AttachController( ObjectClass *pObject );
	void Initializer();
	
	void Tick();

	PushMenu2InitData  DevicesMenuInitData;
	PushMenu2Class    *pDevicesMenu;
	
	PushMenu3InitData  UnitsMenuInitData;
	PushMenu3Class    *pUnitsMenu;
	
	PushButtonInitData MediaButtonInitData;
	PushButtonClass   *pMediaButton;

	std::vector< StorageMediaUserInterface* > vpStorageMediaUser_I;
	std::vector< std::string >                vsOldPaths;
	
	void MakeUnitsMenu();
	
	void DeviceMenuItemSelected_Callback( const char *acDeviceName, int nDeviceIndex );
	void UnitsMenuItemSelected_Callback( const char*, const char *acMediaPath, dword dwData );
	void MediaButtonPushCallback( int nButtonState );
	void MediaButtonPushedInstantCallback();
	
	dword  nSelectedDeviceIndex;
	dword  nSelectedUnit;
	
	int  nPartHandle;
	
	IntegerConstant nC_MediaButtonWidth;
	IntegerConstant nC_MediaButtonWidthAutoAdjust;
};


/////////////////////////////////////////////////////////////////////////////////////////////////////////

#define  SB_DriveButton_CLASS_NAME     "SB_DriveButton"
#define  SB_DriveButton_OBJECT_TYPE    SB_DriveButton_CLASS_NAME

class SB_DriveButton : protected SBarPart_ObjectClass, protected TicksUserClass {
public:
	static const char* const CLASS_NAME;  // "SB_DriveButton"
	static const char* const OBJECT_TYPE; // "SB_DriveButton"

	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;
	

protected:
	virtual ~SB_DriveButton()
	{
		if( pDriveButton ) { pDriveButton->Destroy(); }
	}
	
private:
	SB_DriveButton( ConfigInterface *pConfig_I, const char *acAlias );
	
	void Tick();

	bool PlaceController( WorkfieldOwnerStatusBarClass *pStatusBar );
	bool StorageMediaUser_AttachController( ObjectClass *pObject );

	StorageMediaUserInterface   *pStorageMediaUser_I;
	StorageDriveInterface       *pStorageDrive_I;
	
	DriveButtonInitData ButtonInitData;
	DriveButtonClass   *pDriveButton;

	void DriveButtonPushCallback( int nMountMode )
	{
		if( !nMountMode ) { pConfigI->SetInstantCallback( fastdelegate::MakeDelegate(this, &SB_DriveButton::MountMode0_InstantCallback) ); }
		else              { pConfigI->SetInstantCallback( fastdelegate::MakeDelegate(this, &SB_DriveButton::MountMode1_InstantCallback) ); }
	}
	void MountMode0_InstantCallback() 
	{
		pStorageMediaUser_I->UnMountMedia( nC_UnitNumber );
		pDriveButton->SetMediaPath("");
	}
	void MountMode1_InstantCallback();

	void SetDriveName();

	std::string  sMediaPath;

	IntegerConstant  nC_UnitNumber;
	
	int  nPartHandle;
};

#endif
