//
//  config.h  -     .
//
#ifndef ConfigClassDef
#define ConfigClassDef 1

#include <windows.h>
#include <stdio.h>
#include <string>
#include <map>

#include "..\UtilsLib\main_classes.h"
#include "..\UtilsLib\abstract_model.h"
#include "..\UtilsLib\WFTabs.h"
#include "..\UtilsLib\workfield_class.h"
#include "..\UtilsLib\ini_filer.h"
#include "..\UtilsLib\console_class.h"
#include "..\UtilsLib\status_bar.h"


#define ConfigInterface_CLASS_NAME   "ConfigInterface"
#define ConfigClass_CLASS_NAME       "Config"
#define ConfigClass_OBJECT_TYPE      ConfigClass_CLASS_NAME
#define acConfigClass_SuperType      "config"


#ifndef ConfigStructsDef
#define ConfigStructsDef 1

struct ModuleRecordStruct {
	HINSTANCE           hLib;
	InternalsType      *pModuleInternalsRoot;
	std::string         sFileSpec;
	void_DllFun_pchar	SetModuleLanguage;
	CreateObjectFun     CreateObject;
};

struct ObjectRecordStruct {
	ObjectClass      *pObject;
	std::string       sModuleAlias;
};

struct LinkElementStruct {  //  ElementAlias {SocketType} [SocketAlias] | & < >
	ObjectClass      *pObject;
	std::string       sElementAlias;
	std::string       sSocketType;
	std::string       sSocketAlias;
	int               nDir;  // 0 == IO ; 1 == Output ; -1 = Input ; 7 == Root ; 
};

struct LinksStruct {
	LinkElementStruct               RootElement;
	std::vector<LinkElementStruct>  vElements;
};


class ConfigClass;

struct ConfigSharedData {

	dword  dwHookerPID;

	bool   bNewNumLock;
	dword  uNumLockChangeTC;

	bool   bNewCapsLock;
	dword  uCapsLockChangeTC;

	bool   bNewScrollLock;
	dword  uScrollLockChangeTC;

	dword  uKeyboardChangeTC;

	bool  bSavedNumLock;
	bool  bSavedCapsLock;
	bool  bSavedScrollLock;

	bool  bNumLockState;
	bool  bCapsLockState;
	bool  bScrollLockState;

	bool  bVK_NUMLOCK_pressed;
	bool  bVK_CAPITAL_pressed;
	bool  bVK_SCROLL_pressed;

	int   nActiveMainConfigs;

	HWND  ghWndFocus;
	
	byte  diKeys[256];					     //      DirectInput ( dikCodes.h )
	bool  abDisabled_PassThru_diKeys[256];
	
	LOCK_VAR_TYPE  ghWndFocus_LOCK;
};

#endif

///////////////////////////////////////////////

const dword  ConfigInterface_Generation = 1;
const dword  ConfigInterface_Version    = 1;


class ConfigInterface : public InterfaceClass {
friend unsigned int WINAPI BusThread( void *pArg );
friend class WorkfieldOwnerClass;

public:
	static const char* const CLASS_NAME; // "ConfigInterface"
	
	virtual void Set_dwLoadConfigError_TC( dword dwArg ) { dwLoadConfigError_TC = dwArg; }
	virtual void SetActiveWorkField( WorkField *pWF )=0;
	virtual void SetLanguage( const char *acLanguage )=0;
	virtual void ConStrOut( const char *str = "\n" )=0;
	
	virtual const byte *const Get_diKeys()=0;
	
	virtual void  ReadIniStr(  ObjectClass *pObject, const char *acKey, char *acBuf, const char *acDefVal = "" );
	virtual void  WriteIniStr( ObjectClass *pObject, const char *acKey, const char *acBuf );
	virtual bool  CheckIniStr( ObjectClass *pObject, const char *acKey );

	virtual void  ReadIni_ANSI_Str(  ObjectClass *pObject, const char *acKey, char *acBuf, const char *acDefVal = "" );
	inline  void  WriteIni_ANSI_Str( ObjectClass *pObject, const char *acKey, const char *acVal )
	{
		std::string sSection = pObject->Get_sAlias();
		sSection += ".ini";
		::WriteIni_ANSI_Str( sConfigFile.c_str(), acKey, acVal, sSection.c_str() );
	}
	inline  bool  CheckIni_ANSI_Str( ObjectClass *pObject, const char *acKey )
	{
		return CheckIniStr( pObject, acKey );
	}
	//
	virtual int   ReadIniInteger(  ObjectClass *pObject, const char *acKey, int nDefVal = 0, NumberType Type = DECIM, 
	                               std::vector<const NamedValType>const *pvAllowedVals = NULL );

	virtual int   ReadIniInteger(  ObjectClass *pObject, const char *acKey, int nDefVal, NumberType Type, const int* anPermittedVals, int nPermittedVals );
	virtual int   ReadIniInteger(  ObjectClass *pObject, const char *acKey, int nDefVal, NumberType Type, int nMinVal, int nMaxVal );
	inline  bool  ReadIniBoolean(  ObjectClass *pObject, const char *acKey, int nDefVal )
	{
		return ReadIniInteger( pObject, acKey, nDefVal, DECIM, 0, 1 );
	}
	virtual void  WriteIniInteger( ObjectClass *pObject, const char *acKey, int nVal, NumberType Type = DECIM );
	inline  bool  CheckIniInteger( ObjectClass *pObject, const char *acKey )
	{
		return CheckIniStr( pObject, acKey );
	}
	//
	virtual int   ReadIniRGB(  ObjectClass *pObject, const char *acKey, int nDefVal = 0 );
	inline  void  WriteIniRGB( ObjectClass *pObject, const char *acKey, int nVal  )
	{
		std::string sSection = pObject->Get_sAlias();
		sSection += ".ini";
		::WriteIniRGB( sConfigFile.c_str(), acKey, nVal, sSection.c_str() );
	}
	inline  bool  CheckIniRGB( ObjectClass *pObject, const char *acKey )
	{
		return CheckIniStr( pObject, acKey );
	}

	
	virtual void RunEmulation()=0;
	virtual void StopEmulation()=0;
	virtual bool IsEmulationActive()
	{
		if( pBusI )
		{
			return pBusI->IsEmulThreadActive();
		}
		return false;
	}
	
	virtual bool SetTicksUser( TicksUserInterface *pTicksUserI, bool bMode = ON )
	{
		if( bMode == ON && TicksUserInterface_Generation != pTicksUserI->Get_dwGeneration() ) { return false; }
	
		vpTicksUsers.lock();
		//////////////////////////

		safe_vector< TicksUserInterface* >::iterator it = find (vpTicksUsers.begin(), vpTicksUsers.end(), pTicksUserI );

		if( bMode == ON )
		{
			if( it == vpTicksUsers.end() )
			{
				vpTicksUsers.push_back( pTicksUserI );
			}
		}
		else
		{
			if( it != vpTicksUsers.end() )
			{
				vpTicksUsers.erase( it );
			}
		}
		///////////////////////////
		vpTicksUsers.unlock();
		
		return true;
	}

	virtual bool SetFastTicksUser( FastTicksUserInterface *pFastTicksUserI, int nMode = ON )=0;
	virtual void SetInstantCallback( fastdelegate::FastDelegate0<> Callback )=0;
	
	virtual bool LoadConfig( const char* )=0;
	virtual void CloseConfig()=0;
	virtual bool IsConfigClosed() { return bConfigClosed; }
	virtual void ExitProgram()=0;
	
	
	
	virtual void Disable_diKey_PassThru( byte diKey )=0;
	virtual void Enable_diKey_PassThru( byte diKey )=0;

	virtual void Set_ghWndFocus( HWND hWndFOCUS )=0;
	virtual void Clear_ghWndFocus( HWND hWndFOCUS )=0;
	virtual void Lock_ghWndFocus()=0;
	virtual void UnLock_ghWndFocus()=0;
	virtual HWND Get_ghWndFocus()=0;
	
	virtual void SetNumLock( bool bState )=0;
	virtual void SetScrollLock( bool bState )=0;
	virtual void SetCapsLock( bool bState )=0;

	virtual void SaveNumLock()=0;
	virtual void SaveScrollLock()=0;
	virtual void SaveCapsLock()=0;

	virtual void RestoreNumLock()=0;
	virtual void RestoreScrollLock()=0;
	virtual void RestoreCapsLock()=0;
	
	
//
protected:
	ConfigInterface( ObjectClass *pOwner );
	virtual ~ConfigInterface(){}
	
	virtual WorkField*  CreateWorkField( const char *acTitle, WorkfieldOwnerInterface *pWOI )=0;
	virtual void DestroyWorkField( WorkField *&pWF )
	{
		pTabs->DelWFTab( pWF );
		pWF->Destroy();
		pWF = NULL;
	}

	bool             bConfigClosed;
    std::string      sConfigFile;
	BusInterface    *pBusI;

	WFTabsClass	    *pTabs;
	HWND             hWndTab;
	WorkField       *pActiveWF;
	
	dword            dwLoadConfigError_TC;
	
	safe_vector< TicksUserInterface* >            vpTicksUsers;
	safe_vector< FastTicksUserInterface* >        vpFastTicksUsers;
	
	std::vector< fastdelegate::FastDelegate0<> > *pvdCurrInstantCallbacks;
	std::vector< fastdelegate::FastDelegate0<> >  vdInstantCallbacks1;
	std::vector< fastdelegate::FastDelegate0<> >  vdInstantCallbacks2;
};


//
//
//
class ConfigClass : protected SocketOwnerClass, protected WorkfieldOwnerClass, protected ConfigInterface {

friend LRESULT CALLBACK WndProc1( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );
friend unsigned int WINAPI BusThread( void *pArg );

public:
	static const char* const CLASS_NAME;  // "Config"
	static const char* const OBJECT_TYPE; // "Config"
	
	static ObjectClass*      ObjectConstructor( ConfigInterface*, const char *acAlias );
	static InternalsType     InternalsRoot;

	static InternalsPointersVectorType  vInternalsPointers;
	static InternalsPointersMapType     mInternalsPointers;
	
	static ConfigClass *pCFG;
	static void ConStrOUT( const char *str );
//
protected:
	virtual ~ConfigClass();

	static LRESULT CALLBACK WndProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );

	StateVarType  St_Log;
	void          Log_StateCallback( const char*, int nState )
	{
		if( nState < 0 ) { return; }

		if( nState )
		{
			pConsole->StartLog();
		}
		else
		{
			pConsole->StopLog();
		}
	}
	//
	HANDLE  hEvBusThreadActive;

	
	virtual WorkField*  CreateWorkField( const char *acTitle, WorkfieldOwnerInterface *pWOI );

	virtual bool LoadConfig( const char* );
	virtual void CloseConfig();
	virtual void ExitProgram()
	{
		if( !bMainConfig ) { return pConfigI->ExitProgram(); }
		SendNotifyMessage( hWndMain, WM_CLOSE, 0, 0 );
	}
	
	virtual void RunEmulation();
	virtual void StopEmulation();

	virtual void SetLanguage( const char *acLanguage );
	virtual void ConStrOut( const char *str = "\n" );

	virtual void SetActiveWorkField( WorkField *pWF );
	bool    bParentWF_Active;
	virtual void WF_ActiveStateChanged( bool bActive )
	{
		if( bParentWF_Active == bActive ) { return; }
		bParentWF_Active = bActive;
		if( pTabs ) { pTabs->Activate( bActive ); }
	}
	
	virtual void ThemeChanged();
	virtual void Tick()
	{
		pConsole->FlushLogData();

		for( int i = 0 ; ; i++ )
		{
			TicksUserInterface *pTicksUserI;
			
			vpTicksUsers.lock();
			//////////////////////////
			int nSize = vpTicksUsers.size();
			if( i >= nSize )
			{
				vpTicksUsers.unlock();
				break;
			}
			else
			{
				pTicksUserI = vpTicksUsers[i];
			}
			///////////////////////////
			vpTicksUsers.unlock();

			pTicksUserI->Tick();
		}
	}
	//
	virtual bool SetFastTicksUser( FastTicksUserInterface *pFastTicksUserI, int nMode )
	{
		if( !bMainConfig ) { return pConfigI->SetFastTicksUser( pFastTicksUserI, nMode ); }

		if( nMode != OFF && FastTicksUserInterface_Generation != pFastTicksUserI->Get_dwGeneration() ) { return false; }

		vpFastTicksUsers.lock();
		//////////////////////////

		safe_vector< FastTicksUserInterface* >::iterator it = find (vpFastTicksUsers.begin(), vpFastTicksUsers.end(), pFastTicksUserI );

		if( nMode == ON )
		{
			if( it == vpFastTicksUsers.end() )
			{
				vpFastTicksUsers.push_back( pFastTicksUserI );
			}
		}
		else
		if( nMode == OFF )
		{
			if( it != vpFastTicksUsers.end() )
			{
				vpFastTicksUsers.erase( it );
			}
		}
		///////////////////////////
		vpFastTicksUsers.unlock();
	
		return true;
	}
	virtual void SetInstantCallback( fastdelegate::FastDelegate0<> Callback )
	{
		if( !bMainConfig ) { return pConfigI->SetInstantCallback( Callback ); }

		(*pvdCurrInstantCallbacks).push_back( Callback );
	}
	//

	bool ParseAttach(  const char * const pOrigStr, LinkElementStruct *pElement, const char *pStr );
	bool AttachOblect( const char * const pOrigStr, ObjectClass *pLeftObject, LinkElementStruct *pLeftElement, LinkElementStruct *pRightElement );

    bool ParseLinks( const char *&pStr, LinkElementStruct *pElement = NULL );
   	void PointError( const char *pStr );
	int GetLeftElement( LinkElementStruct *pElement, const char *&pStr );
	int GetElement( LinkElementStruct *pElement, const char *&pStr, bool bRight = false );


	HWND  hWndFrame;
	HWND  hWndStatus;
	
	WFConsole           *pConsole;

	Input_SocketClass    TextInputSocket;
	Input_SocketClass    ByteInputSocket;

	TextOutput_SocketClass   Log;
	Output_SocketClass       ByteOutputSocket;

	void TextReceiver( PVOID pData, SocketClass *pSender ){ ConStrOut( (char*)pData ); }
	void ByteReceiver( PVOID ch, SocketClass *pSender )
	{
		char str[8];
		str[0] = (byte)ch;
		str[1] = '\0';
		TextReceiver( (PVOID)str, pSender );
	}
	void KeyboardReceiver( char ch ) { ByteOutputSocket.Send( (PVOID)ch ); }

	virtual LRESULT CALLBACK Parse_Keyboard_WMs( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
	{
		if( pActiveWF ) { return pActiveWF->Parse_Keyboard_WMs( hWnd, msg, wParam, lParam ); }
		
		if( msg == WM_CHAR ) { KeyboardReceiver( wParam ); }
		return 0;
	}

	LRESULT CALLBACK FrameWndProc( HWND, UINT, WPARAM, LPARAM );

	virtual HWND GetFocusHWND() { return pTabs->GetFocusHWND(); }

	virtual void WF_Tick();

	DWindowInitData  ConInitData;
	Console          *pCon;
	
	virtual LRESULT CALLBACK Parse_WM_MOUSEWHEEL( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
	{
		if( pActiveWF ) { return pActiveWF->Parse_WM_MOUSEWHEEL( hWnd, msg, wParam, lParam ); }
		return 0;
	}
	
	virtual void ResizeToRECT( RECT rc );


	int nMIPS_StatusPartHandle;
	int nMIPS_Limit_StatusPartHandle;
	

    char* Trim( char* );
    int   CutDirSign( char* ); //      '<'  '>',   . 
                               //  [0]:     ; 
                               //            [1]:   : ''; 
                               //           [-1]:   : ''.


	std::map <std::string, ModuleRecordStruct> modules;
	std::map <std::string, ModuleRecordStruct>::iterator pModulesItem;

	std::map <std::string, ObjectRecordStruct> objects;
	std::map <std::string, ObjectRecordStruct>::iterator pObjectsItem;
	

	const char *pOrigStr;
	const char *pStrMax;
    
	
    int nLinksNesting;
	
	char acSection[1024*1024];

private:
	ConfigClass( ConfigInterface *pConfigI, const char *acAlias );
	
	
	virtual const byte *const Get_diKeys() { return pSharedData->diKeys; }
	
	virtual void Disable_diKey_PassThru( byte diKey ) { pSharedData->abDisabled_PassThru_diKeys[ diKey ] = true;  }
	virtual void Enable_diKey_PassThru( byte diKey  ) { pSharedData->abDisabled_PassThru_diKeys[ diKey ] = false; }
	
	virtual void Set_ghWndFocus( HWND hWndFOCUS );
	virtual void Clear_ghWndFocus( HWND hWndFOCUS );
	SharedLockVarType  ghWndFocus_Lock;
	virtual void Lock_ghWndFocus();
	virtual void UnLock_ghWndFocus();
	virtual HWND Get_ghWndFocus();
	
	//
	static HHOOK             ghKeyboardHook; // hook handle
	static LRESULT CALLBACK  KeyboardHook( int nCode, WPARAM wParam, LPARAM lParam );
	static ConfigSharedData *pSharedData2;
	
	dword  dwTicksCount;
	
	dword  dwActiveConfigID;
	bool   bMainConfig;
	bool   bObjectDestroing;
	
	HANDLE hEmulThread;
	
	void CheckKeyboardState( bool bFinal = false );

	virtual void SetNumLock( bool bState );
	virtual void SetScrollLock( bool bState );
	virtual void SetCapsLock( bool bState );

	virtual void SaveNumLock();
	virtual void SaveScrollLock();
	virtual void SaveCapsLock();

	virtual void RestoreNumLock();
	virtual void RestoreScrollLock();
	virtual void RestoreCapsLock();
	
//	virtual void SetNumLockControllingUse(    bool bUse ) { if( bUse ) bNumLockControllingUse    = bUse; }
//	virtual void SetScrollLockControllingUse( bool bUse ) { if( bUse ) bScrollLockControllingUse = bUse; }
//	virtual void SetCapsLockControllingUse(   bool bUse ) { if( bUse ) bCapsLockControllingUse   = bUse; }
	
	bool  bNumLockControllingUse;
	bool  bScrollLockControllingUse;
	bool  bCapsLockControllingUse;

	StringConstant     sC_ConfigFileToLoad;
	IntegerParameter   nP_ExtendedStartLog;

	ConfigSharedData  *pSharedData;
	HANDLE             hMapping;
	
	static unsigned int nObjectsCount;
};

#endif

