//
//  PDP11A16_classes.h  -      PDP11A16
//

#ifndef PDP11A16_classesDefs
#define PDP11A16_classesDefs 1


#include <string>
#include <vector>
#include <map>
#include <deque>
#include <algorithm>
#include <functional>

#include "main_classes.h"
#include "abstract_model.h"
#include "sysmac.h"
#include "FastDelegate.h"
#include "console_class.h"


#pragma warning(disable:4996)


#define PDP11A16 "PDP11A16"

#define WORD(a) (*(unsigned short *)&aMemory[a])
#define BYTE(a) aMemory[a]

typedef fastdelegate::FastDelegate3<IO_ModeType,IO_SizeType,int>  IO_FunDelegateType;

//////////////////////////////////////

class PDP11A16_DeviceClass;
class PDP11A16_CPU_BASE_Class;
class PDP11A16_DeviceInterface;


struct IRQ_RecType {
	word  uVector;
	word  uBR_Line;
	word  uBR_LinePosition;
	bool  bClearedByInit;
	int   nDelay; //     1,   .
	PDP11A16_DeviceInterface *pDevI;
};

struct FastCallbackDataType {
	fastdelegate::FastDelegate0<> Callback;
	int nSteps;
};

struct RegisterRW_MessageDataType {
	std::string     sRegisterName;
	AccessModeType	RW_Mode;
	int             nData;
};

struct MemoryRW_MessageDataType {
	dword           nAddr;
	AccessModeType	RW_Mode;
	int             nBytes;
	byte            *pBuf;
};


struct MemoryAccessDataType {
	byte                       *paMemory;
	ObjectClass*               *pIO_map;
	IO_FunDelegateType         *pIO_map2;
	bool                       *pabRAM;
	bool                       *pabReadAccess;
	bool                       *pabWriteAccess;
	int                        *panAccessLatencyNS;
	int                        *panAccessLatencyClocks;
};

//////////////////////////////

#define  PDP11A16_BusInterface_CLASS_NAME   "PDP11A16_BusInterface"

const dword  PDP11A16_BusInterface_Generation = 1;
const dword  PDP11A16_BusInterface_Version    = 1;


class PDP11A16_BusInterface : public InterfaceClass {
public:
	static const char* const CLASS_NAME; // "PDP11A16_BusInterface"

	virtual void Get_MemoryAccessData( MemoryAccessDataType *pMemData )=0;
	virtual bool Is_V100_DeviceInstalled()=0;
	virtual void Init()=0;

	virtual bool SetIRQ( word uVector, word uBR_Line, word uBR_LinePosition, bool bClearedByInit = true, PDP11A16_DeviceInterface* pDevI = NULL )=0;
	virtual bool ClearIRQ( word uVector )=0;
	virtual void SetFastCallback( fastdelegate::FastDelegate0<> Callback, int nSteps = 4 )=0;

protected:
	virtual ~PDP11A16_BusInterface(){}
	PDP11A16_BusInterface( ObjectClass *pOwner ):InterfaceClass( pOwner, CLASS_NAME, PDP11A16_BusInterface_Generation, PDP11A16_BusInterface_Version ){}
};

/////

struct DeviceComponentRecord
 {
	DeviceComponentRecord(): nLatencyNS(0), pRegsAccessModes(NULL),
	nIO_Regs(0), nIO_RegsBaseAddr(0), nVectors(0), nVectorsBaseAddr(0){}
	
	bool  bComponentDisabled; //         ,
	                          //            .
	int   nIO_Regs;           //    -
	word  nIO_RegsBaseAddr;   //      -
	int   nVectors;           //    
	word  nVectorsBaseAddr;   //     
	int   nLatencyNS;         //         (   TRUE SPEED ).
	
	AccessModeType   RegsAccessMode;
	AccessModeType *pRegsAccessModes;
	
	std::string sDescription;
};

//////////////////////////////

class PDP11A16_CPU_BASE_Interface;

#define  PDP11A16_DeviceInterface_CLASS_NAME       "PDP11A16_DeviceInterface"

const dword  PDP11A16_DeviceInterface_Generation = 1;
const dword  PDP11A16_DeviceInterface_Version    = 1;


class PDP11A16_DeviceInterface : public InterfaceClass {
public:
	static const char* const CLASS_NAME; // "PDP11A16_DeviceInterface"

	virtual void IO_fun( IO_ModeType IO_Mode, IO_SizeType IO_Size, int nIO_Addr ){}
	virtual void IRQ_Recived( word wAddr )=0;
	virtual std::vector<DeviceComponentRecord>& vComponentsRef()=0;
	virtual void Init(){}

	virtual IO_FunDelegateType  Get_IO_funDelegate()
	{
		return fastdelegate::MakeDelegate(this, &PDP11A16_DeviceInterface::IO_fun);
	}
//
protected:
	virtual ~PDP11A16_DeviceInterface(){}
	PDP11A16_DeviceInterface( ObjectClass *pOwner ):InterfaceClass( pOwner, CLASS_NAME, PDP11A16_DeviceInterface_Generation, PDP11A16_DeviceInterface_Version ){} //      
};

//////////////////////////////

#define  PDP11A16_CPU_BASE_Interface_CLASS_NAME   "PDP11A16_CPU_BASE_Interface"

const dword  PDP11A16_CPU_BASE_Interface_Generation = 1;
const dword  PDP11A16_CPU_BASE_Interface_Version    = 1;


class PDP11A16_CPU_BASE_Interface : public InterfaceClass {
public:
	static const char* const CLASS_NAME; // "PDP11A16_CPU_BASE_Interface"

	virtual PDP11A16_DeviceInterface* Get_p16DevI()=0;
	
	virtual bool   SetIRQ( word uVector, word uBR_Line, word uBR_LinePosition, bool bClearedByInit = true, PDP11A16_DeviceInterface* pDevI = NULL )=0;
	virtual bool   ClearIRQ( word uVector )=0;
	virtual void   SetFastCallback( fastdelegate::FastDelegate0<> Callback, int nSteps = 4 )=0;

protected:
	virtual ~PDP11A16_CPU_BASE_Interface(){}
	PDP11A16_CPU_BASE_Interface( ObjectClass *pOwner ):InterfaceClass( pOwner, CLASS_NAME, PDP11A16_CPU_BASE_Interface_Generation, PDP11A16_CPU_BASE_Interface_Version ){}
};

//////////////////////////////

class PDP11A16_DeviceClass : virtual protected AbstractDeviceClass, protected PDP11A16_DeviceInterface {
protected:
	virtual ~PDP11A16_DeviceClass(){}
	PDP11A16_DeviceClass();

	PDP11A16_BusInterface  *p16BusI;

	std::vector<DeviceComponentRecord>           vComponents;
	virtual std::vector<DeviceComponentRecord>&  vComponentsRef() { return vComponents; }

	byte                 *aMemory;
	ObjectClass*         *IO_map;
	IO_FunDelegateType   *IO_map2;
	bool                 *abRAM;
	bool                 *abReadAccess;
	bool                 *abWriteAccess;
	int                  *anAccessLatencyNS;
	int                  *anAccessLatencyClocks;


	virtual void IRQ_Recived( word wAddr )       { vIRQ[ wAddr ] = false; }
	virtual void PreEmulation_Report( const char *acDeviceDescription );

	std::map<word,bool> vIRQ;

private:
	int  LinkCheck( SocketClass*, SocketClass* );
	void LinkDone(  SocketClass*, SocketClass* );
};

//////////////////


//
//  PDP11A16_CPU_BASE_Class -- Message Macros
//

#define RegisterRW_MES( pCPU, sRegisterName, RW_Mode, nData ) \
{ \
	RegisterRW_MessageDataType RegRW = { sRegisterName, RW_Mode, nData }; \
	MessageType                Mes   = { "RegisterRW", (qword)&RegRW }; \
	pCPU->Message( &Mes ); \
}


class PDP11A16_CPU_BASE_Class : protected AbstractCPU_Class, protected PDP11A16_DeviceClass, protected PDP11A16_CPU_BASE_Interface {
protected:
	virtual ~PDP11A16_CPU_BASE_Class(){}
	PDP11A16_CPU_BASE_Class( bool bTrueSpeedSupported );

	TextOutput_SocketClass    DebugOut;

	virtual PDP11A16_DeviceInterface* Get_p16DevI() { return (PDP11A16_DeviceInterface*)this; };
	

	StateVarType   St_HALT;
	StateVarType   St_Timer;
	

	virtual void Init();
	virtual void PrepareEmulation();
	virtual void InitComponents(){};
	
	IO_ModeType  Last_IO_Mode;
	IO_SizeType  Last_IO_Size;
	int          nLast_IO_Addr;

	inline void READ_DONE()
	{
		if( nLast_IO_Addr >= 0 ) 
		{
			IO_map2[ nLast_IO_Addr ]( IO_READ_DONE, Last_IO_Size, nLast_IO_Addr );
		}
	}
	
	virtual void Make_IO()
	{
		if( nLast_IO_Addr >= 0 )
		{
			IO_map2[ nLast_IO_Addr ]( Last_IO_Mode, Last_IO_Size, nLast_IO_Addr );
		}
	}

//	
	int      nLostTrapsCount;
	
	
	Console *pCon;
	DWindowInitData  ConInitData;
};


#endif