#ifndef ConsoleClassDefs
#define ConsoleClassDefs 1
//
//   console_class.h
//

#include <time.h>
#include "sysmac.h"
#include "logger.h"
#include "window_init_data.h"
#include "DBufClass.h"

//////////////////////////////////
typedef unsigned char        byte;
typedef unsigned short       word;
typedef unsigned long       dword;
typedef unsigned long long  qword;
//////////////////////////////////


#define DEFAULT_CONSOLE_BACKGROUND  RGB(0xB5,0xC7,0xC9)


//  ,  GetLogFileState();  //

// enum { LOG_OFF, LOG_ON, LOG_SUSPENDED };


//
//
//
//#ifndef MaxLastStrLenDef
//#define MaxLastStrLenDef 1
//const int MAX_LAST_STR_LEN = 512;
//#endif

//
//
//
const int CONSOLE_DEF_BUF_SIZE = 1024*24;

class Console {

public:

	static const char* const CLASS_NAME; // = "Console";

	void SetFocus( void );
	void CharOut( char );
	void StrOut( const char* );
	void StatusStrOut( char* );
	void TitleStrOut( char* );

	void ClrStr( void );

	Console( DWindowInitData*, unsigned long = CONSOLE_DEF_BUF_SIZE );

	virtual ~Console();

	LRESULT CALLBACK MasterWndProc( HWND, UINT, WPARAM, LPARAM );
	LRESULT CALLBACK ConsoleWndProc( HWND, UINT, WPARAM, LPARAM );

	void Quit();

	void  SetLogFile( const char *acPath );
	void  StartLog()
	{
		if( nLogState == LOG_OFF )
		{
			hLog = LoggerOpen( sLogFile.c_str() );
		}
		nLogState = LOG_ON;
	}
	void  StopLog() { nLogState = LOG_SUSPENDED; FlushLogData(true); }
	void  CloseLog(){ nLogState = LOG_OFF; if( hLog ) { sLogData += pcLastStr; FlushLogData(true); LoggerClose( hLog ); hLog = 0; } }
	void  FlushLogData( bool bForce = false )
	{
		dwTC = GetTickCount();
		
		if( sLogData.size() )
		{
			if( bForce || dwTC - dwLastLogFlushTC > 2000 )
			{
				dwLastLogFlushTC = dwTC;
				
				ConLock.Lock();
				/////////////////
				LoggerStrOut( hLog, sLogData.c_str() );
				sLogData.clear();
				/////////////////
				ConLock.UnLock();
			}
		}
	}
	//
	void Resize( int, int, int, int ); //    (child) !
	void WF_Tick();

	LRESULT CALLBACK Parse_WM_MOUSEWHEEL( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );

protected:

	Console(){};  //      

	DBufClass     *TextBuf;

	void          ShowScreen( HWND, HDC );
	void          InitInternals(void);
	void          InitLogFileData(void);

	char          acObjectName[128];

	bool          bTicks;
	dword         dwTC;

	DWindowInitData*    lpInitData;

	std::string  sLogFile;
	std::string  sLogData;
	int          hLog;
	int          nLogState;

	// private:

	int  nStrOutPos;

	static int ObjectsCount( bool = false );

	bool bObjectDestroing;
	bool bWindowDestroed;
	
	dword dwLastLogFlushTC;

	HDC  hDCmain;
	HDC  hDCcon;
	HDC  hDC_EOF_pic;

	HBITMAP hEOF_bitmap;
	UINT    hTimer;

	HGDIOBJ hWorkWindowBackgrownd;
	//  HGDIOBJ hMainWindowBackgrownd;

	int     hConsoleLogFile;
	void    LogFileStrOut( char* );

	HWND    hWndStatus;
	HWND    hWndCon;
	HWND    hWndMain;
	HWND    hWndFrame;

	int     nMaxStrLen;
	int     nMaxMaxStrLen;

	bool    nHBarDisabled;

	unsigned long uLastScrollUnixTime;

	bool    bIsStatusWindow;
	bool    bIsEmbedded;
	int     nEmbeddedBorderWidth;
	int     nEmbeddedStyleNum;
	int     nFrameDecorWidth;
	bool    bEmbeddedTitle;
	bool    bHideScrollBars;

	int     nStatusHeight;

	int     nUpdateScreenFlag;
	bool    bBufShifted;

	int     nCtrRFlag;

	char    *pcTopLeft;
	char    *pcTopLeftOrig;
	char    *pcTextCurPos;
	char    *pcLastChar;
	char    *pcLastStr;

	char    *pcPrevLastChar;

	int     nTextCol0;
	int     nTextCol;

	int     nTextRow0;
	int     nTextRow0Orig;
	int     nTextRow;

	int     nRow;
	int     nCol;
	int     nTextRowCount;

	int     nConsoleX0;
	int     nConsoleY0;

	int     nCursorFlag;
	int     nBlinkFlag;
	int     nBlinkSusupendFlag;
	int     nBlinkStopFlag;
	int     nBlinkFix_On_Flag;
	int     nBlinkFix_OFF_Flag;

	int     nStartFlag;
	int     nPaintEofCharFlag;

	int     nScreenRows;
	int     nScreenCols;

	int     nBufShiftFlag;


	long int      nLastConPrint;
	long int      nLastCursorBlink;

	long int      nLastFocusTime;

	int           nWindowMinimizedFlag;

	int           nScrollerPage, nScrollerMax, nScrollerPos;
	int           nScrollerWidth, nScrollerStrMax, nScrollerXPos;
	int           nHBarState;

	int           nLButtonFlag;

	TEXTMETRIC    metric;
	HFONT         fontX;
	RECT          con_rect;

	int           nFontDX;
	int           nFontDY;

	dword         dwLastShowScreenTC;


	HWND          CreateStatusWindow( HWND, int = 77 );
	//  HWND          CreateHScrollBar( HWND );
	HWND          CreateWorkWindow( HWND, DWindowInitData*, int = 78 );

	//char     str[80];

	void ConPrintChar( char );
	void ConCursorDisable( HWND );
	void ConCursorEnable( HWND );
	void ConCursorOFF( HWND hWnd );
	void ConCursorON( HWND hWnd );
	void PaintEofChar( HDC, int, int );
	void ShiftBuf( char* );

	void UpdateScrollBar( int, int, int );
	void UpdateHBar( int, int, int );
	void MoveScrollBox( HWND hWnd, HDC hDC, int );
	void ScrollLineDown( HWND hWnd, HDC hDC );
	void ScrollLineUp( HWND hWnd, HDC hDC );
	void ScrollPageUp( HWND hWnd, HDC hDC );
	void ScrollPageDown( HWND hWnd, HDC hDC );

	//  void ShowScreen( HWND, HDC );
	void BufInsertChar( char, int );

	void ConCursorInvert0( HWND );
	void ConCursorInvert( HWND );
	void ConCursorPaint( HWND );
	void ConCursorBlinkOff( HWND );
	void ConCursorBlinkOn( HWND );
	void ConCursorBlink( HWND );

	void RestoreTruePos( void );
	void UpdateTruePos( void );
	void CheckPointers( void );

	LockVarType  ConLock;
};

//
//  Monitor -  
//
class Monitor : private Console {

public:

  static const char* const CLASS_NAME; // = "Monitor";

  Monitor( DWindowInitData*    lpInitData,
           unsigned long       lBufSize = CONSOLE_DEF_BUF_SIZE );

  ~Monitor();

  void IdentOut( char*, char* );

  void TitleStrOut( char *str )
  {
    Console::TitleStrOut( str );
  };

  void Resize( int nX0, int nY0, int nDX, int nDY )
  {
    Console::Resize( nX0, nY0, nDX, nDY );
  };

  void WF_Tick()
  {
    Console::WF_Tick();
  };

private:

  void SearchIdentStr( char* );
  void MakeIdentStr( char*, char*, char* );
  void ReplaceStr( char * );
  void ShiftVals( void );

  int  nMaxIdentLen;

  static int ObjectsCount( bool = false );

};

#endif

