//
//  main_classes.h  -     API .
//

#ifndef main_classesDefs
#define main_classesDefs 1

#include "sysmac.h"
#include "FastDelegate.h"
#include "console_class.h"

#pragma warning(disable:4996)


////////////////

#define  acStream_SuperType  "stream"
#define  acBus_SuperType     "bus"
#define  acDevice_SuperType  "device"
#define  acCPU_SuperType     "cpu"
#define  acTool_SuperType    "tool"
#define  acLine_SuperType    "line"

#define  acStateLineType     "StateLine"

////////////////

const int  OFF          = 0;
const int  ON           = 1;
const int  CHECK        = 2;
const int  DISABLED_ON  = BIT_31 | ON;
const int  DISABLED_OFF = BIT_31 | OFF;
const int  NOT_EXIST    = -1;


////////////////////////////////


enum ObjectDataType  : int { D_STATE = 0, D_INTEGER_PARAM = 1, D_STRING_PARAM = 2, D_INTEGER_CONST = 3, D_STRING_CONST = 4, D_LAST };


////////////////////////////////

extern const char* acIncompatible_Generations_of_class( const char *acClassName );

class GlobalClass {
public:
	virtual dword  Get_dwGeneration(){ return dwGeneration; }
	virtual dword  Get_dwVersion()   { return dwVersion;    }
	
protected:
	virtual ~GlobalClass(){}
	GlobalClass(): dwGeneration(0), dwVersion(0) {}

	GlobalClass( dword uGeneration, dword uVersion ):
	dwGeneration( uGeneration ),
	dwVersion( uVersion )
	{
	}
private:
	dword dwGeneration;
	dword dwVersion;
};

////////////////////////////////

class WFObject;

class ObjectClass;
class SocketClass;
class SocketOwnerClass;
class ConfigInterface;
class SocketOwnerInterface;

typedef std::vector<SocketClass*> SocketsPointersVector;


enum ParameterMode  : int {  SLAVE_MODE = -1, STANDARD_MODE = 0, BASIC_MODE = 1, STROBE_MODE = 2  };
enum AccessModeType : int {  NONE = 0, READ = 1, WRITE = 2, READnWRITE = 3  };
enum IO_ModeType    : int {  IO_NONE,   IO_ERR,   IO_READ,   IO_WRITE,   IO_READ_DONE  };
enum IO_SizeType    : int {  IO_SIZE_ERR,   IO_BYTE,   IO_WORD  };
enum LineModeType   : int {  NO_LINE = 0, INPUT_LINE = 1, OUTPUT_LINE = 2, INPUT_and_OUTPUT_LINES = 3, IO_LINE = 4 };
enum MediaTypeConst : int {  MEDIA_TYPE_NONE, MEDIA_TYPE_DSK, MEDIA_TYPE_TRK };

//////////////////////////////////////////////////////////

extern void                     ConStrOut( const char *str = "\n" );
extern HINSTANCE                ghInstance;
extern const char*              StartWinSockUse();
extern void                     StopWinSockUse();
struct ProgramDescriptorsType;
extern ProgramDescriptorsType  *pProgramDescriptors;



typedef  ObjectClass*  ObjectConstructorType( ConfigInterface*, const char* );
typedef  ObjectClass*  (__cdecl* CreateObjectFun)( ConfigInterface *pConfigI, const char *acType, const char *acAlias );


struct InternalsType;

typedef std::map<std::string,InternalsType*> InternalsPointersMapType;
typedef std::vector<InternalsType*>          InternalsPointersVectorType;

struct InternalsType {
	const char* const                acName;
	const char* const                acDescription;
	const char* const                acDataType;
	const char* const                acDataFormat;
	InternalsPointersMapType * const pInternalsPointersMap;
	qword       const                qwData;
};

extern InternalsPointersVectorType vpUtilsLibInternals;

typedef InternalsType*  (__cdecl* ModuleInitFun)( ProgramDescriptorsType  *pProgram_Descriptors );

void            InitInternalsPointersMaps( InternalsType *pInternal );
InternalsType*  LoadModule( ProgramDescriptorsType *pProgramDescriptors, const char *acModuleFile );

class AddInternal {
public:
	AddInternal( InternalsPointersVectorType *pvInternalsPointers, InternalsType *pInternal )
	{
		(*pvInternalsPointers).push_back( pInternal );
	}
};

#define SetComponentBuildInternal(component_name) InternalsType component_name##Build = { #component_name "#build", #component_name " Component Build", "char*", "%s", NULL, (qword)__DATE__ " " __TIME__ }; AddInternal component_name##Build_( &vpUtilsLibInternals, & component_name##Build );


class WorkField;
struct ProgramDescriptorsType {
	HWND			hWndMain;
	WorkField      *pWF;
	void(*pConStrOut)(const char*);
	InternalsType  *pProgramInternalsRoot;
	void           (__cdecl* AddModuleInternals)( const char *acModuleFile, InternalsType *pModuleInternalsRoot );
	const char     *acProgramPath;
	const char     *acProgramName;
};

//////////////////////////////////////////////////////////

class InterfaceClass : public GlobalClass {
public:
	virtual	ObjectClass*  Get_pOwner() { return pOwner; }
	
protected:
	virtual ~InterfaceClass(){}
	InterfaceClass( ObjectClass *pOwner, const char *acName, dword uGeneration, dword uVersion );
	ObjectClass *pOwner;
};

//////////////////////////////////////////////////////////

typedef fastdelegate::FastDelegate2<SocketClass*,SocketClass*,int>  LinkCheckControllerDelegateType;
typedef fastdelegate::FastDelegate2<SocketClass*,SocketClass*>      LinkDoneControllerDelegateType;

/////

const dword  SocketClass_Generation = 1;
const dword  SocketClass_Version    = 1;

class SocketClass : public GlobalClass, protected NoCopy {

friend class SocketOwnerClass;
friend class StateVarType;

public:
	static const char* const CLASS_NAME; // "Socket"

	virtual ObjectClass*        Get_pOwner()           const{ return pOwner;           }
	virtual int                 Get_nDir()             const{ return nDir;             }
	virtual std::string const&  Get_sSocketAlias()     const{ return sSocketAlias;     }
	virtual std::string const&  Get_sSocketType()      const{ return sSocketType;      }
	virtual std::string const&  Get_sSocketSuperType() const{ return sSocketSuperType; }
	virtual int                 Linked()               const{ return vpLinkedSockets.size(); }
	virtual void                Reset()                     { vpLinkedSockets.clear(); }
	
	virtual void AddLinkCheckController( LinkCheckControllerDelegateType   Controller )
	{
		vdLinkCheckControllers.push_back( Controller );
	}

	virtual void AddLinkDoneController( LinkDoneControllerDelegateType   Controller )
	{
		vdLinkDoneControllers.push_back( Controller );
	}
	
	virtual void AddCommandSupport( const char *acCommand, fastdelegate::FastDelegate2<PVOID,SocketClass*> CommandProcessor )
	{
		mCommandProcessors[ acCommand ] = CommandProcessor;
		vsSupportedCommands.push_back( acCommand );
	}
	virtual void AddReplySupport(   const char *acReply,   fastdelegate::FastDelegate2<PVOID,SocketClass*> ReplyProcessor )
	{
		mReplyProcessors[ acReply ] = ReplyProcessor;
		vsSupportedReplys.push_back( acReply );
	}
	//
	fastdelegate::FastDelegate2< PVOID,SocketClass* >                Receive;
	fastdelegate::FastDelegate3< const char*,PVOID,SocketClass* >    CommandReceive;  // SocketClass* -   ,  .
	fastdelegate::FastDelegate3< const char*,PVOID,SocketClass* >    ReplyReceive;    // SocketClass* -   ,  .

	virtual void SendCommand( const char *acCommand, PVOID pData = NULL );
	virtual void SendReply(   const char *acReply,   PVOID pData, SocketClass *pReceiver = NULL );  // pReceiver -   ,    
	                                                                                                //             (  -     ).
	
protected:
	virtual ~SocketClass()
	{
		for( int i = 0 ; i < vpLinkedSockets.size() ; i++ )
		{
			vpLinkedSockets[i]->UnLink( this );
		}
	}
	//
	//  acSocketAlias ( " " )  -     ,
	//                    .         
	//                    -.
	//
	//  nDIR -  : 0 - , 1 - , -1 - .
	//
	SocketClass( const char *acSocketSuperType, const char *acSocketType, const char *acSocketAlias, int nDIR, SocketOwnerInterface *pSOI );

	ObjectClass           *pOwner;
	SocketOwnerInterface  *pSOI;
	std::string            sSocketAlias;
	std::string            sSocketSuperType;
	std::string            sSocketType;
	int                    nDir;
	
	virtual int  Link( SocketClass* );           //  -   .
	virtual int  LinkCheck( SocketClass* );      //      .
	virtual void LinkDone( SocketClass* );       //      .
	virtual void UnLink( SocketClass* );         //   .
	
	virtual void CacheFlush(){}

	SocketsPointersVector vpLinkedSockets;
	
	std::map< std::string, fastdelegate::FastDelegate2<PVOID,SocketClass*> > mCommandProcessors;
	std::vector< std::string >                                               vsSupportedCommands;

	std::map< std::string, fastdelegate::FastDelegate2<PVOID,SocketClass*> > mReplyProcessors;
	std::vector< std::string >                                               vsSupportedReplys;

private:
	void CommandReceiver( const char* acCommand, PVOID pData, SocketClass *pSender );
	void ReplyReceiver(   const char* acCommand, PVOID pData, SocketClass *pSender );

	void CommandsReporter( PVOID, SocketClass *pReceiver );
	void ReplysReporter(   PVOID, SocketClass *pReceiver );

	std::vector< LinkCheckControllerDelegateType >  vdLinkCheckControllers;
	std::vector< LinkDoneControllerDelegateType  >  vdLinkDoneControllers;

	static unsigned int nObjectsCount;
};

////////

class IO_SocketClass: public SocketClass {
public:
	static const char* const CLASS_NAME; // "IO_Socket"

	IO_SocketClass( const char *acSuperType, const char *acType, const char *acAlias, 
                    SocketOwnerInterface *pSOI, fastdelegate::FastDelegate2< PVOID,SocketClass* > Receiver );
	virtual ~IO_SocketClass(){}

	virtual void Send(PVOID);
	inline  void SendByte( byte ch ) { return Send( (PVOID)ch ); }
	inline  void operator()(PVOID pArg ){ Send(pArg); }
	
private:
  static unsigned int nObjectsCount;
};


////////

class Input_SocketClass: public SocketClass {
public:
	static const char* const CLASS_NAME; // "Input_Socket"

	Input_SocketClass( const char *acSuperType, const char *acType, const char *acAlias, 
                       SocketOwnerInterface *pSOI, fastdelegate::FastDelegate2< PVOID,SocketClass* > Receiver );
	virtual ~Input_SocketClass(){}

private:
  static unsigned int nObjectsCount;
};

////////

class Output_SocketClass: public SocketClass {
public:
	static const char* const CLASS_NAME; // "Output_Socket"

	Output_SocketClass( const char *acSuperType, const char *acType, const char *acAlias, 
                        SocketOwnerInterface *pSOI );
	virtual ~Output_SocketClass(){}

	virtual void Send(PVOID);
	inline  void SendByte( byte ch ) { return Send( (PVOID)ch ); }
	inline  void operator()(PVOID pArg ){ Send(pArg); }
	
private:
  static unsigned int nObjectsCount;
};

////////

class TextOutput_SocketClass: public SocketClass {
public:
	static const char* const CLASS_NAME; // "TextOutput_Socket"

	TextOutput_SocketClass( const char *acSuperType, const char *acType, const char *acAlias, 
                            SocketOwnerInterface *pSOI, unsigned int nCacheMaxSize );
	virtual ~TextOutput_SocketClass(){}

	virtual void Send(PVOID);
	inline  void operator()( const char *pArg ){ Send( (PVOID)pArg ); }

protected:

	unsigned int nOutputCacheMaxSize;
	
	mutable std::deque<std::string> OutputCache;
	
	virtual void CacheFlush();
	
private:
  static unsigned int nObjectsCount;
};


//////////////////////////////////////////////////////////

struct StateCallbacksVectorElement {
	std::string                                   sAlias;
	fastdelegate::FastDelegate2<const char*,int>  Callback;
};

typedef std::map< std::string, fastdelegate::FastDelegate2<const char*,int> >         StateCallbacksMap;
typedef std::deque< StateCallbacksVectorElement >                                     StateCallbacksDeque;

typedef std::map< std::string, fastdelegate::FastDelegate2<const char*,int> >         IntParamCallbacksMap;
typedef std::map< std::string, fastdelegate::FastDelegate2<const char*,const char*> > StrParamCallbacksMap;


const dword  StateVarType_Generation = 1;
const dword  StateVarType_Version    = 1;


class StateVarType : public GlobalClass, protected NoCopy {
public:
	virtual ~StateVarType();
	StateVarType( const char *acName, bool nDefVAL, ObjectClass *pOWNER, LineModeType LineMode = NO_LINE, ParameterMode  SMode = STANDARD_MODE );

	virtual ObjectClass* Get_pOwner() { return pOwner; }

	virtual void Load();
	virtual void Save();
	virtual void Reset();

	virtual int  Init( int nVal )
	{
		nDefVal = OFF;
		if( nVal < 0 && nVal != DISABLED_OFF ) { nVal = DISABLED_ON; nDefVal = ON; }
		if( nVal > 0 )                         { nVal = ON;          nDefVal = ON; }
		if( nState == nVal ) { return nState; }
		nState = nVal;
		SignalCallbacks();
		return nState;
	}
	virtual int  Set(  int nVal )
	{
		if( nState < 0 ) { return nState; }
		if( nVal   < 0 || nState == nVal ) { return nState; }
		if( nVal != 0 ) { nState = 1; }
		else            { nState = 0; }
		SignalCallbacks();
		return nState;
	}
	virtual int  Get()const{ return nState; }
	virtual bool SetCallback( const char *acCallbackAlias, fastdelegate::FastDelegate2<const char*,int> Callback, bool bPlaceFirst = false, bool bSendSignal = true );
	virtual void SignalCallbacks();

	inline  void operator=( bool bState ) { Set( bState ); }
	inline  void operator=( const StateVarType& StateVar ) { Set( StateVar() ); }

	inline  int  operator()(void)const
	{
		if( nState < 0 ) { return nDefVal; }
		return nState;
	}
	inline  operator int() const{ return operator()(); }
	
	std::string const  sName;

protected:
	int                nState;
	int                nLoadedState;
	int                nDefVal;
	bool               bLoaded;
	bool               bSaveChanges;
	LineModeType       LineMode;
	bool               bLineBusy;
	ParameterMode      Mode;

	ObjectClass         *pOwner;
	Input_SocketClass   *pInputSocket;
	Output_SocketClass  *pOutputSocket;

//	StateCallbacksMap  mCallbacks;
	StateCallbacksDeque  vCallbacks;
	
private:
	void StateReceiver( PVOID uState, SocketClass *pSender );
	void StateCallback( const char*,int );
	void OutputLinkDoneController( SocketClass *pRemoteSocket, SocketClass *pOwnSocket );
};

//////////////////////////////////////////////////////////

struct NamedValType {
	int            nVal;
	const char    *acValAlias;
};


const dword  IntegerParameter_Generation = 1;
const dword  IntegerParameter_Version    = 1;


class IntegerParameter : public GlobalClass, protected NoCopy {
friend class IntegerParametersClass;
public:
	virtual ~IntegerParameter();
	IntegerParameter( const char *acName, const char *acGroup, ObjectClass *pOWNER, int nDefVAL, int nMinVAL, int nMaxVAL, 
	                  std::vector<const NamedValType>const *pvAllowedVals = NULL, NumberType NType = DECIM, ParameterMode PMode = STANDARD_MODE );

	IntegerParameter( const char *acName, const char *acGroup, ObjectClass *pOWNER, int nDefVAL, std::vector<const int>const *pvnAllowedValues,
	                  NumberType NType = DECIM, ParameterMode PMode = STANDARD_MODE );

	virtual ObjectClass* Get_pOwner(){ return pOwner; }

	virtual void SetCallback( const char *acCallbackAlias, fastdelegate::FastDelegate2<const char*,int> Callback, bool bSendSignal = true );
	virtual void Load();
	virtual void Reset();
	virtual void Set( int nVal );
	virtual int  Get()const{ return nVal; }

	virtual std::vector<const int> const* Get_pvnAllowedValues()
	{
		return &vnAllowedValues;
	}
	

	inline  int  operator()(void)const{	return Get(); }
	inline       operator int()  const{ return Get(); }

	inline  void operator=(int nNewVal) { Set( nNewVal ); }
	inline  void operator=(const IntegerParameter& ParamVar ) { Set( ParamVar() ); }


protected:	
	std::string const sName;
	std::string const sGroup;

	int               nVal;
	int               nLoadedVal;
	bool              bLoaded;
	ParameterMode     Mode;
	NumberType        Type;
	int const         nDefVal;
	int               nMinVal;
	int               nMaxVal;

	ObjectClass             *pOwner;

	std::vector<const int>            vnAllowedValues;
	std::vector<const NamedValType>   vAllowedVals;
	
	IntParamCallbacksMap     mCallbacks;
};

//////////////////////////////////////////////////////////

const dword  IntegerConstant_Generation = 1;
const dword  IntegerConstant_Version    = 1;


class IntegerConstant : public GlobalClass, protected NoCopy {
public:
	virtual ~IntegerConstant(){}
	IntegerConstant( const char *acName, const char *acGroup, ObjectClass *pOWNER, int nDefVAL, int nMinVAL, int nMaxVAL, 
	                 std::vector<const NamedValType>const *pvAllowedVals = NULL, NumberType NType = DECIM, ParameterMode CMode = STANDARD_MODE, bool bDisabled = false );
	
	IntegerConstant( const char *acName, const char *acGroup, ObjectClass *pOWNER, int nDefVAL, std::vector<const int>const *pvnAllowedValues, 
	                 NumberType NType = DECIM, ParameterMode CMode = STANDARD_MODE, bool bDisabled = false );

	virtual int  Get()const{ return nVal; }

	inline  int  operator()(void)const{	return Get(); }
	inline       operator int()  const{ return Get(); }

protected:
	void CommonConstructor();
	
	std::string const sName;
	std::string const sGroup;

	int               nVal;
	ParameterMode     Mode;
	NumberType        Type;
	int               nDefVal;
	int               nMinVal;
	int               nMaxVal;

	ObjectClass      *pOwner;

	std::vector<const int>            vnAllowedValues;
	std::vector<const NamedValType>   vAllowedVals;
};

//////////////////////////////////////////////////////////

const dword  StringParameter_Generation = 1;
const dword  StringParameter_Version    = 1;


class StringParameter : public GlobalClass, protected NoCopy {

friend class StringParametersClass;
//         - (.. ,   - ).
friend inline bool operator ==( StringParameter& ParamVar1, StringParameter& ParamVar2 ){ return ( ParamVar1.sVal == ParamVar2.sVal ); }
friend inline bool operator !=( StringParameter& ParamVar1, StringParameter& ParamVar2 ){ return ( ParamVar1.sVal != ParamVar2.sVal ); }
friend inline bool operator ==( std::string sStr, StringParameter& ParamVar ){ return ( sStr == ParamVar.sVal ); }
friend inline bool operator ==( StringParameter& ParamVar, std::string sStr ){ return ( sStr == ParamVar.sVal ); }
friend inline bool operator ==( StringParameter& ParamVar, const char *pStr ){ return ( ParamVar.sVal == pStr ); }
friend inline bool operator ==( const char *pStr, StringParameter& ParamVar ){ return ( ParamVar.sVal == pStr ); }
friend inline bool operator !=( std::string sStr, StringParameter& ParamVar ){ return ( sStr != ParamVar.sVal ); }
friend inline bool operator !=( StringParameter& ParamVar, std::string sStr ){ return ( sStr != ParamVar.sVal ); }
friend inline bool operator !=( StringParameter& ParamVar, const char *pStr ){ return ( ParamVar.sVal != pStr ); }
friend inline bool operator !=( const char *pStr, StringParameter& ParamVar ){ return ( ParamVar.sVal != pStr ); }

public:
	virtual ~StringParameter();
	StringParameter( const char *acName, const char *acGroup, ObjectClass *pOWNER, const char *acDefVAL = "", ParameterMode PMode = STANDARD_MODE );

	virtual ObjectClass* Get_pOwner(){ return pOwner; }

	virtual void SetCallback( const char *acCallbackAlias, fastdelegate::FastDelegate2<const char*,const char*> Callback, bool bSendSignal = true );
	virtual void Load();
	virtual void Reset();
	virtual void Set( const char *acNewVal );
	virtual const char* Get()const{ return sVal.c_str(); }

	inline  const char* operator()(void)  const{ return Get(); }
	inline  operator const char*        ()const{ return Get(); }

	inline  void operator=( const char *acNewVal ) { Set( acNewVal ); }
	inline  void operator=( const std::string sNewVal ) { Set( sNewVal.c_str() ); }
	inline  void operator=( const StringParameter& ParamVar ) { Set( ParamVar() ); }

protected:
	std::string const sName;
	std::string const sGroup;
	std::string const sDefVal;

	std::string       sVal;
	std::string       sLoadedVal;
	bool              bLoaded;
	ParameterMode     Mode;

	ObjectClass           *pOwner;
	
	StrParamCallbacksMap   mCallbacks;
};

//////////////////////////////////////////////////////////

const dword  StringConstant_Generation = 1;
const dword  StringConstant_Version    = 1;


class StringConstant : public GlobalClass, protected NoCopy {

//         - (.. ,   - ).
friend inline bool operator ==( StringConstant& Const1, StringConstant& Const2 ){ return ( Const1.sVal == Const2.sVal ); }
friend inline bool operator !=( StringConstant& Const1, StringConstant& Const2 ){ return ( Const1.sVal != Const2.sVal ); }
friend inline bool operator ==( std::string sStr, StringConstant& Const ){ return ( sStr == Const.sVal ); }
friend inline bool operator ==( StringConstant& Const, std::string sStr ){ return ( sStr == Const.sVal ); }
friend inline bool operator ==( StringConstant& Const, const char *pStr ){ return ( Const.sVal == pStr ); }
friend inline bool operator ==( const char *pStr, StringConstant& Const ){ return ( Const.sVal == pStr ); }
friend inline bool operator !=( std::string sStr, StringConstant& Const ){ return ( sStr != Const.sVal ); }
friend inline bool operator !=( StringConstant& Const, std::string sStr ){ return ( sStr != Const.sVal ); }
friend inline bool operator !=( StringConstant& Const, const char *pStr ){ return ( Const.sVal != pStr ); }
friend inline bool operator !=( const char *pStr, StringConstant& Const ){ return ( Const.sVal != pStr ); }

public:
	virtual ~StringConstant(){}
	StringConstant( const char *acName, const char *acGroup, ObjectClass *pOwner, const char *acDefVAL = "", ParameterMode CMode = STANDARD_MODE, bool bDisabled = false );

	virtual const char* Get()const{ return sVal.c_str(); }

	inline  const char* operator()(void)  const{ return sVal.c_str(); }
	inline  operator const char*        ()const{ return sVal.c_str(); }

protected:
	std::string const sName;
	std::string const sGroup;
	std::string const sDefVal;

	std::string       sVal;
	ParameterMode     Mode;
};

//////////////////////////////////////////////////////////

struct MessageType {
	std::string const  sMessage;
	qword              qwData;
};

struct AttachPointRecordType {
	std::string                                      sName;
	std::string                                      sType;
	fastdelegate::FastDelegate1<ObjectClass*,bool>   AttachController;
};


typedef fastdelegate::FastDelegate1<MessageType*,bool>           MessageControllerDelegateType;
typedef std::map< std::string, MessageControllerDelegateType >   MessageControllersMap;

typedef std::map< std::string, PVOID >  InterfacesMap;
typedef std::map< std::string, PVOID >  DataPointersMap;


////////////////////////////////////////
////////////  ObjectClass  /////////////
////////////////////////////////////////
class SocketOwnerInterface;
class WorkfieldOwnerInterface;

const dword  ObjectClass_Generation = 1;
const dword  ObjectClass_Version    = 1;


class ObjectClass : public GlobalClass {
friend int PASCAL WinMain( HINSTANCE, HINSTANCE, LPSTR, int );
friend class WorkfieldOwnerStatusBarClass;
friend class ConfigClass;
friend class SocketClass;
friend class InterfaceClass;
friend class IntegerConstant;
friend class StringConstant;
friend class StateVarType;
friend class IntegerParameter;
friend class StringParameter;

public:
	//==========       ObjectClass ==========//

	static const char* const CLASS_NAME; // "ObjectClass"

	virtual void   Destroy()              { delete this; }
protected: //  Get_nCreationSuccess()  0 -     .
	virtual int    Get_nCreationSuccess() { bObjectCreated = true;  return nCreationSuccess;   }
public:

	virtual std::string const&           Get_sAlias()             { return sAlias;             }
	virtual std::string const&           Get_sType()              { return sType;              }
	virtual std::string const&           Get_sSuperType()         { return sSuperType;         }
	virtual InternalsType*               Get_pInternalsRoot()     { return pInternalsRoot;     }
	virtual SocketOwnerInterface*        pSOI()                   { return NULL;               }
	virtual WorkfieldOwnerInterface*     pWOI()                   { return NULL;               }
protected:
	virtual void SetLastError( const char *acError ){ sLastError = acError; }
public:
	virtual const char*                  Get_acLastError()        { return sLastError.c_str(); }
	virtual ConfigInterface*             Get_pConfigI()           { return pConfigI;           }

	virtual const DataPointersMap*       Get_pDataPointersMap( ObjectDataType nDataType ) 
	{
		if( nDataType < D_LAST ){ return &vmDataPointers[ nDataType ]; }
		return NULL;
	}
	
	virtual PVOID  Get_pData( ObjectDataType nDataType, const char *acDataName )
	{
		if( nDataType >= D_LAST ){ return NULL; }
		DataPointersMap& DP_Map = vmDataPointers[ nDataType ];
		DataPointersMap::iterator it = DP_Map.find( acDataName );
		if( it == DP_Map.end() ){ return NULL; }
		return it->second;				
	}

	virtual PVOID  Get_pInterface( const char *acInterface ) 
	{
		InterfacesMap::iterator it = mInterfaces.find( acInterface );
		if( it == mInterfaces.end() ){ return NULL; }
		return it->second;				
	}

	virtual void   PostLinkInit()           
	{
		const DataPointersMap           *pDP_Map;
		DataPointersMap::const_iterator  it;
		DataPointersMap::const_iterator  end;

//		States.Load();
		pDP_Map = &vmDataPointers[ D_STATE ];
		for( it = pDP_Map->begin(), end = pDP_Map->end() ; it != end ; it++ )
		{
			((StateVarType*)(it->second))->Load();
		}
		
		
//		IntegerParameters.Load();
		pDP_Map = &vmDataPointers[ D_INTEGER_PARAM ];
		for( it = pDP_Map->begin(), end = pDP_Map->end() ; it != end ; it++ )
		{
			((IntegerParameter*)(it->second))->Load();
		}


//		StringParameters.Load();

		pDP_Map = &vmDataPointers[ D_STRING_PARAM ];
		for( it = pDP_Map->begin(), end = pDP_Map->end() ; it != end ; it++ )
		{
			((StringParameter*)(it->second))->Load();
		}

		for( int i = 0, nSize = vdPostLinkInitializers.size() ; i < nSize ; i++ )
		{
			vdPostLinkInitializers[i]();
		}
	}

	virtual bool   Message( MessageType *pMes )
	{
		MessageControllersMap::iterator it = mMessageControllers.find( pMes->sMessage );
		if( it != mMessageControllers.end() )
		{
			return it->second( pMes );
		}
		return false;
	}
	virtual bool   AttachObject( ObjectClass *pObject, const char *acName = "", const char *acType = "" );
	
	virtual void   SetLanguage( const char *acLanguage ){}
	virtual void   ConStrOut( const char *str );
	
	//==========       ObjectClass ==========//

protected:
//	std::string sObjectName;

	virtual ~ObjectClass(){}
	ObjectClass( const char *acBadClass )
	{
		char str[512];
		sprintf(str,"%s: ObjectClass --- Init ERROR ---", acBadClass );
		ErrLog(str);
	}
	ObjectClass( ConfigInterface *pConfig_I, const char *acSuperType, const char *acType, const char *acAlias, InternalsType *pInternalsROOT );

	virtual void AddAttachPoint( AttachPointRecordType& AttachPointRecord )
	{
		for( int nSize = vAttachPoints.size() , i = 0 ; i < nSize ; i++ )
		{
			if( AttachPointRecord.sName == vAttachPoints[i].sName &&
			    AttachPointRecord.sType == vAttachPoints[i].sType )
			{
				char str[512];
				sprintf(str,"PROGRAMM ERROR: Duplicated attach point in object [%s]: Alias[%s] ; Type[%s] ", 
				            sAlias.c_str(), AttachPointRecord.sName, AttachPointRecord.sType );
				ErrLog(str);
				break;
			}
		}
		vAttachPoints.push_back( AttachPointRecord );
	}
	//
	virtual void AddInterface( const char *acInterface, PVOID pInterface ) 
	{
		mInterfaces[acInterface] = pInterface;
		vsInterfaces.push_back( acInterface );
	}
	virtual void AddMessage( const char *acMessage, MessageControllerDelegateType  MessageController ) 
	{
		mMessageControllers[acMessage] = MessageController;
		vsMessages.push_back( acMessage );
	}
	//
	std::string         sAlias;
	std::string const   sType;
	std::string const   sSuperType;
	InternalsType      *pInternalsRoot;
	std::string         sLastError;
	int                 nCreationSuccess;
	bool                bObjectCreated;

	ConfigInterface *pConfigI;

	std::vector< fastdelegate::FastDelegate0<> > vdPostLinkInitializers;

	std::vector< DataPointersMap >         vmDataPointers;

private:
	MessageControllersMap                  mMessageControllers;
	std::vector< std::string >             vsMessages;
	InterfacesMap                          mInterfaces;
	std::vector< std::string >             vsInterfaces;
	std::vector< AttachPointRecordType >   vAttachPoints;
};


////////////////////////////////////////////////
//////////////  TicksUserClass  ////////////////
////////////////////////////////////////////////

const dword  TicksUserInterface_Generation = 1;
const dword  TicksUserInterface_Version    = 1;



class TicksUserInterface : public InterfaceClass {
public:
	static const char* const CLASS_NAME; // "TicksUserInterface"
	
	virtual void  Tick()=0;
protected:
	virtual ~TicksUserInterface(){}
	TicksUserInterface( ObjectClass *pOwner ):InterfaceClass( pOwner, CLASS_NAME, TicksUserInterface_Generation, TicksUserInterface_Version ){}
};

class TicksUserClass : virtual protected ObjectClass, protected TicksUserInterface {
protected:
	virtual ~TicksUserClass();
	TicksUserClass();
};

////////////////////////////////////////////////////
//////////////  FastTicksUserClass  ////////////////
////////////////////////////////////////////////////

const dword  FastTicksUserInterface_Generation = 1;
const dword  FastTicksUserInterface_Version    = 1;


class FastTicksUserInterface : public InterfaceClass {
public:
	static const char* const CLASS_NAME; // "FastTicksUserInterface"
	
	virtual void  FastTick()=0;

protected:
	virtual ~FastTicksUserInterface(){}
	FastTicksUserInterface( ObjectClass *pOwner ):InterfaceClass( pOwner, CLASS_NAME, FastTicksUserInterface_Generation, FastTicksUserInterface_Version ){}
};

class FastTicksUserClass : virtual protected ObjectClass, protected FastTicksUserInterface {
protected:
	virtual ~FastTicksUserClass();
	FastTicksUserClass();

	void  StartFastTicks();
	void  StopFastTicks();
	int   FastTicksState() { return nTicksState; }
	
private:
	int   nTicksState;
};


////////////////////////////////////////////////
////////////  WorkfieldOwnerClass  /////////////
////////////////////////////////////////////////

const dword  WorkfieldOwnerInterface_Generation = 1;
const dword  WorkfieldOwnerInterface_Version    = 1;


class WorkfieldOwnerInterface : public InterfaceClass {
public:
	static const char* const CLASS_NAME; // "WorkfieldOwnerInterface"
	
	virtual void   ThemeChanged()=0;
	virtual void   ResizeWindows( bool bReResize_WF = false )=0;
	virtual void   WF_Tick()=0;
	virtual HWND   GetFocusHWND()=0;
	virtual HWND   GetHWND_WF() { return hWnd_WF; }
	virtual HWND   Get_hWndMain() { return hWndMain; }
	virtual bool   IsVisible() { return IsWindowVisible( hWnd_WF ); }
	virtual void   WF_ActiveStateChanged( bool bActive ){}

	virtual LRESULT CALLBACK Parse_WM_MOUSEWHEEL( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam ){ return 0; }
	virtual LRESULT CALLBACK Parse_Keyboard_WMs( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam ) { return 0; }

	virtual WorkfieldOwnerStatusBarClass* Get_pStatusBar()  { return pStatusBar; }
	virtual WorkfieldOwnerStatusBarClass* Get_pControlBar() { return pControlBar; }

protected:
	virtual void ResizeToRECT( RECT rc ){}

	virtual ~WorkfieldOwnerInterface(){}
	WorkfieldOwnerInterface( ObjectClass *pOwner ):
	InterfaceClass( pOwner, CLASS_NAME, WorkfieldOwnerInterface_Generation, WorkfieldOwnerInterface_Version ),
	pWF(NULL),
	hWnd_WF(NULL),
	hWndMain(NULL),
	pControlBar(NULL),
	pStatusBar(NULL)
	{}  

	WorkField *pWF;
	HWND       hWnd_WF;
	HWND       hWndMain;
	RECT       rcWF;
	
	std::string  sTabTitle;

	WorkfieldOwnerStatusBarClass  *pStatusBar;
	WorkfieldOwnerStatusBarClass  *pControlBar;
};

class WorkfieldOwnerClass : virtual protected ObjectClass, protected TicksUserClass, protected WorkfieldOwnerInterface {
protected:
	virtual ~WorkfieldOwnerClass();
	WorkfieldOwnerClass();

	virtual WorkfieldOwnerInterface*   pWOI() { return (WorkfieldOwnerInterface*)this; }


	virtual void   ThemeChanged();
	virtual void   ResizeWindows( bool bReResize_WF = false );
	
	StringConstant  sC_TabTitle;
};


/////////////////////////////////////////////
////////////  SocketOwnerClass  /////////////
/////////////////////////////////////////////

const dword  SocketOwnerInterface_Generation = 1;
const dword  SocketOwnerInterface_Version    = 1;


class SocketOwnerInterface : public InterfaceClass {
friend class SocketClass;
public:
	static const char* const CLASS_NAME; // "SocketOwnerInterface"
	
	virtual int                 Get_nChannelWays()  { return nChannelWays;  }
	virtual std::string const&  Get_sArchitecture() { return sArchitecture; }
	
	virtual int  Link( SocketClass*, const char *acAlias = "" )=0;  //  .

	virtual const SocketsPointersVector& Get_vpSockets() { return vpSockets; }
	

protected:
	virtual ~SocketOwnerInterface(){}
	SocketOwnerInterface( ObjectClass *pOwner ):InterfaceClass( pOwner, CLASS_NAME, SocketOwnerInterface_Generation, SocketOwnerInterface_Version ){}
	
	SocketsPointersVector   vpSockets;

	std::string    sArchitecture;
	int            nChannelWays;             // 0 -   ; 1 -   ; 2 -  
};

/////

class SocketOwnerClass: virtual protected ObjectClass, protected SocketOwnerInterface {
friend class SocketClass;

protected:
	virtual ~SocketOwnerClass(){}
	SocketOwnerClass( const char *acArchitecture, int nWays = 0 );

	virtual SocketOwnerInterface*  pSOI() { return (SocketOwnerInterface*)this; }

	virtual int  Link( SocketClass*, const char *acAlias = "" );

	TextOutput_SocketClass  SysOut;
};


//////////////////////////////////////

class ToolClass : protected SocketOwnerClass {
public:
	static const char* const CLASS_NAME; // "tool"
	
protected:
	virtual ~ToolClass(){}
	ToolClass();
};


//////////////////////////////////////

const dword  BusInterface_Generation = 1;
const dword  BusInterface_Version    = 1;


class BusInterface : public InterfaceClass {
public:
	static const char* const CLASS_NAME; // "BusInterface"

	virtual bool          PrepareEmulation(){ return true; }
	virtual dword         EmulThread()=0;
	virtual bool          IsEmulThreadActive()=0;
	virtual void          StopEmulation()=0;

protected:
	virtual ~BusInterface(){}
	BusInterface( ObjectClass *pOwner ):InterfaceClass( pOwner, CLASS_NAME, BusInterface_Generation, BusInterface_Version ){}
};

/////

class BusClass : protected SocketOwnerClass, protected BusInterface {
public:
	static const char* const CLASS_NAME; // "bus"

protected:
	virtual ~BusClass(){}
	BusClass( const char *acArchitecture );

	IO_SocketClass   BusSocket;

	StateVarType     St_Pause;

private:
	void BusReceiver( PVOID, SocketClass* ){}
};

//////////////////////////////////////

class BusUserClass : protected SocketOwnerClass {
protected:
	virtual ~BusUserClass(){}
	BusUserClass( const char *acArchitecture );

	IO_SocketClass   BusSocket;
	ObjectClass     *pBus;
	
private:
	void BusReceiver( PVOID, SocketClass* ){}

	int  LinkCheck( SocketClass*, SocketClass* );
	void LinkDone( SocketClass*, SocketClass* );
};

////////////////////////////////////////////////////////////////////////////

struct StorageMediaDescriptor {
	std::string sPath;
	int  nMountingPercent; //  -1 --> Not mounted
};

const dword  StorageMediaUserInterface_Generation = 1;
const dword  StorageMediaUserInterface_Version    = 1;


class StorageMediaUserInterface : public InterfaceClass {
public:
	static const char* const CLASS_NAME; // "StorageMediaUserInterface"

	virtual bool MountMedia( int nUnit, char *acPath, bool bSavePath = true )=0;
	virtual void UnMountMedia( int nUnit )=0;

	virtual int MediaMountingPercent( int nUnit )
	{
		if( nUnit < vUnits.size() ) { return vUnits[nUnit].nMountingPercent; }
		return -1;
	};
	virtual const std::vector< StorageMediaDescriptor >& Get_vUnits() { return vUnits; }
	virtual const char* Get_acDeviceName() { return acDeviceName; }
	virtual const int   Get_nUnitsTotal()  { return nUnitsTotal; }

	virtual const char* Get_acDefaultMediaFileExt() { return acDefaultMediaFileExt; }

protected:
	virtual ~StorageMediaUserInterface(){}
	StorageMediaUserInterface( ObjectClass *pOwner, const char *acDeviceNAME, const int nUnitsTOTAL, const char *acDefExt ):
	InterfaceClass( pOwner, CLASS_NAME, StorageMediaUserInterface_Generation, StorageMediaUserInterface_Version ),
	acDeviceName( acDeviceNAME ),
	nUnitsTotal( nUnitsTOTAL ),
	acDefaultMediaFileExt( acDefExt )
	{}

	std::vector< StorageMediaDescriptor > vUnits;
	const char* const              acDeviceName;
	const int                      nUnitsTotal;
	const char* const              acDefaultMediaFileExt;
};


////////////////////////////////////////////////////////////////////////////

const dword  StorageDriveInterface_Generation = 1;
const dword  StorageDriveInterface_Version    = 1;


class StorageDriveInterface : public InterfaceClass {
public:
	static const char* const CLASS_NAME; // "StorageDriveInterface"

	virtual bool Get_bMotorON(   int nUnit )=0;
	virtual bool Get_bSelected(  int nUnit )=0;
	virtual bool Get_bProtected( int nUnit )=0;
	virtual int  Get_nTrack(     int nUnit )=0;

protected:
	virtual ~StorageDriveInterface(){}
	StorageDriveInterface( ObjectClass *pOwner ):
	InterfaceClass( pOwner, CLASS_NAME, StorageDriveInterface_Generation, StorageDriveInterface_Version )
	{}

};


#endif