//
//   status_bar.h
//

#ifndef StatusWindowClassDefs
#define StatusWindowClassDefs 1

#include "sysmac.h"
#include <stdio.h>

#include "controls.h"

const int STATUS_HEIGHT = 20;



struct StatusPartDataType {
	bool            bUsed;
	int             nPart;         //     -1,   .
	int             nPosition;     //     ( 1,2,3 .. )   ( -1,-2,-3 .. )
	int             nWidth;
	int             nTextWidth;
	int             nPriority;     //            .
	bool            bHidden;       //   . 
	std::string     sText;

	std::vector< ControlsClass* > vpControls;
	
	StatusPartDataType()
	{
		memset( this, 0, (byte*)&sText-(byte*)&bUsed );
	}
};


/////////////////////////////////////////////////////

const dword  WorkfieldOwnerStatusBarClass_Generation = 1;
const dword  WorkfieldOwnerStatusBarClass_Version    = 1;


class WorkfieldOwnerStatusBarClass : public GlobalClass {
public:
	WorkfieldOwnerStatusBarClass( WorkfieldOwnerInterface *pWOI, HWND hWNDParent, 
	                              const char *acName = NULL, int nStatusHeight = 0, bool bSizeGRIP = false, bool bTop = false );

	virtual const char*  Get_acLastError() { return sLastError.c_str(); }

	virtual void Show( int nCmdShow = SW_SHOW ) 
	{
		St_StatusBar.Set( nCmdShow );
	}
	inline  void Hide()	{ Show( SW_HIDE ); }

	virtual	ObjectClass* Get_pOwner() { return pWOI->Get_pOwner(); }
		
	virtual int  AddNamedPart( const char *acPartName, int nTextWidth, int nPosition = 1, int nPriority = 0, ControlsClass *pControl = NULL );

	static LRESULT NewStatusWindowProc( HWND hWndStatus, UINT uMsg, WPARAM wParam, LPARAM lParam );	
	
	virtual void Destroy() { delete this; }
	virtual HWND Get_HWND(){ return hWndStatus; }

	//           AddPart   nPartHandle  0  .
	virtual int  AddPart( int nTextWidth, int nPosition = 1, int nPriority = 0, ControlsClass *pControl = NULL, bool bHidden = false );
	virtual bool AddControl( int nPartHandle, ControlsClass *pControl, int nPos = -1 );
	virtual bool RemoveControl( int nPartHandle, ControlsClass *pControl );
	virtual bool ChangeControl( int nPartHandle, ControlsClass *pOldControl, ControlsClass *pNewControl );
	virtual void StrOut( const char *acText, int nPartHandle = 1 );
	
	virtual int  Height()const;
	virtual void Resize();
	virtual void SetMainPartMinimalWidth( int nWidth );
	virtual void ShowPart( int nPartHandle, bool bShow = true  )
	{
		StatusPartDataType& PartData = vStatusPartsData[ nPartHandle ];
		PartData.bHidden = !bShow;
		if( !bShow ) { PartData.nPart = -1; }
		Resize();
	}
	inline  void HidePart( int nPartHandle )
	{
		ShowPart( nPartHandle, false );
	}
	virtual void ThemeChanged();
	virtual void Reset()
	{
		Reset00();
		St_StatusBar.Reset();
		St_StatusBar.SetCallback( "WorkfieldOwnerStatusBarClass", fastdelegate::MakeDelegate(this, &WorkfieldOwnerStatusBarClass::StatusBar_StateCallback) );
	}
//

protected:
	virtual ~WorkfieldOwnerStatusBarClass();
	
	virtual void Reset00();

	std::vector< StatusPartDataType >    vStatusPartsData;
	std::vector< int >                   vStatusPartsLayout;
	
	HWND hWndParent;
	HWND hWndStatus;
	
	HFONT hStatusFont;
	
	bool bObjectDestroing;
	bool bWindowDestroed;
	bool bVisible;
	bool bReCreating;
	
	bool bSizeGrip;
	bool bTop;
	bool bThemedStyle;
	int  nStatusMinHeight;
	int  nMin_MainPartWidth;
	int  nSeparatorWidth;

	COLORREF  dwBackground;
	
	bool HideLowPriorityParts( int nSpaceToFree );
	bool UnHideLowPriorityParts( int nSpaceToFill );


	StateVarType     St_StatusBar;
	void             StatusBar_StateCallback( const char*, int nState );

	bool StatusBarPart_AttachController( ObjectClass *pObject );	
	
	WorkfieldOwnerInterface  *pWOI;
	
	std::string               sName;
	AttachPointRecordType     AttachPointRecord;
	
	std::string               sLastError;

	void LoadNamedPartData( const char *acPartName, StatusPartDataType *pPartData );	
};

/////////////////////////////////////////////////////
#endif