//
//   stub_architecture.cpp - stub architecture with CPU, keyboard & video for modular API.
//

#include "stub_architecture.h"
#include "UtilsLib\ByteStream.h"
#include "UtilsLib\dikCodes.h"

//          Video,
static DWORD_ALIGN volatile LONG pActiveBus;
//      ,   "" .


const char* const stub_BusInterface::CLASS_NAME = stub_BusInterface_CLASS_NAME;

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

const char* const stub_BusClass::CLASS_NAME  = stub_BusClass_CLASS_NAME;
const char* const stub_BusClass::OBJECT_TYPE = stub_BusClass_OBJECT_TYPE;

unsigned int stub_BusClass::nObjectsCount = 0;

InternalsPointersVectorType  stub_BusClass::vInternalsPointers;
InternalsPointersMapType     stub_BusClass::mInternalsPointers;

InternalsType  stub_BusClass::InternalsRoot = { OBJECT_TYPE, "Object Internals Root", "InternalsPointersVectorType*", "0x%X", &mInternalsPointers, (qword)&vInternalsPointers };

InternalsType  stub_Bus_ObjectConstructor = { "ObjectConstructor", NULL, "ObjectConstructorType*", "0x%X", NULL, (qword)stub_BusClass::ObjectConstructor };
AddInternal    i_stub_Bus_ObjectConstructor( &stub_BusClass::vInternalsPointers, &stub_Bus_ObjectConstructor );

ObjectClass* stub_BusClass::ObjectConstructor( ConfigInterface *pConfigI, const char *acAlias )
{
	return new stub_BusClass( pConfigI, acAlias );
}

/////////////////////


stub_BusClass::stub_BusClass( ConfigInterface *pConfigI, const char *acAlias ):
ObjectClass( pConfigI, acBus_SuperType, OBJECT_TYPE, acAlias, &InternalsRoot ),
stub_BusInterface( (ObjectClass*)this ),
AbstractBusClass( stub_architecture )
{
	if( !nCreationSuccess ) { return; }

	nObjectsCount++;
	
	BusSocket.AddLinkDoneController( fastdelegate::MakeDelegate(this, &stub_BusClass::LinkDone)  );
	
	memset( vpPortOwners, NULL, sizeof(vpPortOwners) );


SysOut("stub_Bus -- created !!!\n");	
}

void stub_BusClass::LinkDone( SocketClass* pSocket, SocketClass* )
{   
	ObjectClass *pObject = pSocket->Get_pOwner();
	stub_DeviceInterface *pDev_I = (stub_DeviceInterface*) pObject->Get_pInterface( stub_DeviceInterface_CLASS_NAME );
	
	if( pDev_I )
	{
		vpInstalledDevices.push_back( pObject );
	
		const PortsVectorType &vPorts = *pDev_I->Get_vPorts();
		for( int nSize = vPorts.size(), i = 0 ; i < nSize ; i++ )
		{
			vpPortOwners[ vPorts[i] ] = pDev_I;
		}
	}
}


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

// enum stub_CPU_command_codes  : byte { NOP, TEST_PORT, JUMP_NOT_READY, READ_PORT, WRITE_PORT, INC, JUMP };
int  stub_CPU_command_clocks[] =       { 5,   12,        18,             24,        32,         8,   14   };

byte stub_CPU_ROM[] = {
	NOP,					// 00
	TEST_PORT,				// 01
	'A',					// 02
	JUMP_NOT_READY,			// 03
	0x01,					// 04
	READ_PORT,				// 05
	'A',					// 06
	NOP,					// 07
	TEST_PORT,				// 08
	'B',					// 09
	JUMP_NOT_READY,			// 0A
	0x08,					// 0B
	WRITE_PORT,				// 0C
	'B',					// 0D

	JUMP,                   // 0E
	0x18,                   // 0F

//	NOP,                    // 0E
//	NOP,                    // 0F
	
	TEST_PORT,				// 11
	'C',					// 12
	JUMP_NOT_READY,			// 13
	0x00,					// 14
	READ_PORT,				// 15
	'C',					// 16
	INC,					// 17
	TEST_PORT,				// 18
	'D',					// 19
	JUMP_NOT_READY,			// 1A
	0x18,					// 1B
	WRITE_PORT,				// 1C
	'D',					// 1D
	
	JUMP,					// 1E
	0x00					// 1F
};

const char* const stub_CPU_Class::CLASS_NAME  = stub_CPU_Class_CLASS_NAME;
const char* const stub_CPU_Class::OBJECT_TYPE = stub_CPU_Class_OBJECT_TYPE;

unsigned int stub_CPU_Class::nObjectsCount = 0;

InternalsPointersVectorType  stub_CPU_Class::vInternalsPointers;
InternalsPointersMapType     stub_CPU_Class::mInternalsPointers;

InternalsType  stub_CPU_Class::InternalsRoot = { OBJECT_TYPE, "Object Internals Root", "InternalsPointersVectorType*", "0x%X", &mInternalsPointers, (qword)&vInternalsPointers };

InternalsType  stub_CPU_ObjectConstructor = { "ObjectConstructor", NULL, "ObjectConstructorType*", "0x%X", NULL, (qword)stub_CPU_Class::ObjectConstructor };
AddInternal    i_stub_CPU_ObjectConstructor( &stub_CPU_Class::vInternalsPointers, &stub_CPU_ObjectConstructor );

ObjectClass* stub_CPU_Class::ObjectConstructor( ConfigInterface *pConfigI, const char *acAlias )
{
	return new stub_CPU_Class( pConfigI, acAlias );
}

///////////////////////////

stub_CPU_Class::stub_CPU_Class( ConfigInterface *pConfigI, const char *acAlias ):
ObjectClass( pConfigI, acCPU_SuperType, OBJECT_TYPE, acAlias, &InternalsRoot ),
AbstractCPU_Class( stub_architecture, true ),
AbstractDeviceClass( stub_architecture ),
uProgramCounter(0)
{
	if( !nCreationSuccess ) { return; }

	nObjectsCount++;

	BusSocket.AddLinkDoneController( fastdelegate::MakeDelegate(this, &stub_CPU_Class::LinkDone)  );

SysOut("stub_CPU -- created !!!\n");	
}

void stub_CPU_Class::LinkDone( SocketClass* pSocket, SocketClass* )
{   
	ObjectClass *pObject = pSocket->Get_pOwner();
	stub_BusInterface *pStubBus_I = (stub_BusInterface*) pObject->Get_pInterface( stub_BusInterface_CLASS_NAME );
	
	if( pStubBus_I )
	{
		pvpPortOwners = pStubBus_I->Get_vpPortOwners();
	}
}

void stub_CPU_Class::CPU_Steps( dword nSteps ) 
{
	nStepCount = 0;
	
	while( nSteps-- )
	{
		nStepCount++;

		byte nCmd = stub_CPU_ROM[ uProgramCounter ];

		dwCurrentStepsClocksCount += stub_CPU_command_clocks[ nCmd ];
		
		switch( nCmd )
		{
			case NOP:
			{
//SysOut("NOP\n");
			}
			break;
			
			case TEST_PORT:
			{
// SysOut("TEST_PORT\n");
				uProgramCounter++;
				byte uPortNum = stub_CPU_ROM[ uProgramCounter ];
char str[512];
sprintf(str,"'%c'\n", uPortNum );
// SysOut(str);
				stub_DeviceInterface *pPortOwner = pvpPortOwners[ uPortNum ];

				bPortReadyFlag = false;
				if( !pPortOwner ) { break; }
				bPortReadyFlag = pPortOwner->TestPort( uPortNum );
			}
			break;
			
			case JUMP_NOT_READY:
			{
// SysOut("JUMP_NOT_READY\n");
				uProgramCounter++;
				byte uNewProgramCounter = stub_CPU_ROM[ uProgramCounter ];
char str[512];
sprintf(str,"0x%02X\n", uNewProgramCounter );
// SysOut(str);
				if( !bPortReadyFlag )
				{
					uProgramCounter = uNewProgramCounter;
					continue;
				}
			}
			break;
			
			case READ_PORT:
			{
//SysOut("READ_PORT\n");
				uProgramCounter++;
				byte uPortNum = stub_CPU_ROM[ uProgramCounter ];
				stub_DeviceInterface *pPortOwner = pvpPortOwners[ uPortNum ];
				if( !pPortOwner ) { break; }
				uAccumulator = pPortOwner->ReadPort( uPortNum );
char str[512];
sprintf(str,"'%c':%c\n", uPortNum, uAccumulator );
//SysOut(str);
			}
			break;
			
			case WRITE_PORT:
			{
//SysOut("WRITE_PORT\n");
				uProgramCounter++;
				byte uPortNum = stub_CPU_ROM[ uProgramCounter ];
				stub_DeviceInterface *pPortOwner = pvpPortOwners[ uPortNum ];
				if( !pPortOwner ) { break; }
				pPortOwner->WritePort( uPortNum, uAccumulator );
char str[512];
sprintf(str,"'%c':%c\n", uPortNum, uAccumulator );
//SysOut(str);
			}
			break;
			
			case INC:
			{
//SysOut("INC\n");
				uAccumulator++;
			}
			break;
			
			case JUMP:
			{
//SysOut("JUMP\n");
				uProgramCounter++;
				byte uNewProgramCounter = stub_CPU_ROM[ uProgramCounter ];
				uProgramCounter = uNewProgramCounter;
char str[512];
sprintf(str,"0x%02X\n", uNewProgramCounter );
//SysOut(str);
				continue;
			}
			break;
		}
		
		uProgramCounter++;
		
		if( uProgramCounter >= sizeof( stub_CPU_ROM ) )
		{
			uProgramCounter = 0;
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

const char* const stub_DeviceInterface::CLASS_NAME = stub_DeviceInterface_CLASS_NAME;

///////////////////////////////////////////////////////////////////////////////

const char* const stub_Keyboard_Class::CLASS_NAME  = stub_Keyboard_Class_CLASS_NAME;
const char* const stub_Keyboard_Class::OBJECT_TYPE = stub_Keyboard_Class_OBJECT_TYPE;

unsigned int stub_Keyboard_Class::nObjectsCount = 0;

InternalsPointersVectorType  stub_Keyboard_Class::vInternalsPointers;
InternalsPointersMapType     stub_Keyboard_Class::mInternalsPointers;

InternalsType  stub_Keyboard_Class::InternalsRoot = { OBJECT_TYPE, "Object Internals Root", "InternalsPointersVectorType*", "0x%X", &mInternalsPointers, (qword)&vInternalsPointers };

InternalsType  stub_Keyboard_ObjectConstructor = { "ObjectConstructor", NULL, "ObjectConstructorType*", "0x%X", NULL, (qword)stub_Keyboard_Class::ObjectConstructor };
AddInternal    i_stub_Keyboard_ObjectConstructor( &stub_Keyboard_Class::vInternalsPointers, &stub_Keyboard_ObjectConstructor );

ObjectClass* stub_Keyboard_Class::ObjectConstructor( ConfigInterface *pConfigI, const char *acAlias )
{
	return new stub_Keyboard_Class( pConfigI, acAlias );
}

///////////////////////////

stub_Keyboard_Class::stub_Keyboard_Class( ConfigInterface *pConfigI, const char *acAlias ):
ObjectClass( pConfigI, acDevice_SuperType, OBJECT_TYPE, acAlias, &InternalsRoot ),
AbstractDeviceClass( stub_architecture ),
stub_DeviceInterface( (ObjectClass*)this ),
bExit(false)
{
	if( !nCreationSuccess ) { return; }

	nObjectsCount++;
	
	vPorts.push_back('A');    //   'A'

	aKeys = pConfigI->Get_diKeys();

	memset( abKeysPrev, false, sizeof(abKeysPrev) );
	

SysOut("stub_Keyboard -- created !!!\n");	
}


void stub_Keyboard_Class::InstantCallback()
{
	if( bExit ) { return; }

	if( (dword)pActiveBus == (dword)pBus )
	{//       Video .
		for( int diKey = 0 ; diKey < 256 ; diKey++ )
		{
			bool bKeyPressed     = aKeys[ diKey ];
			bool bKeyPressedPrev = abKeysPrev[ diKey ];
			
			if( bKeyPressed )
			{
				if( !bKeyPressedPrev )
				{
	char str[512];
	sprintf(str,"diKey[0x%X]\n", diKey );
	//SysOut(str);
					sInputBuf.lock();
					/////////////////
					sInputBuf += diKey;
					/////////////////
					sInputBuf.unlock();
					
					abKeysPrev[diKey] = true;
				}
			}
			else
			{
				abKeysPrev[diKey] = false;
			}
		}
	}

	pConfigI->SetInstantCallback( fastdelegate::MakeDelegate(this, &stub_Keyboard_Class::InstantCallback) );
}

byte stub_Keyboard_Class::ReadPort( byte uPort ) 
{
	if( uPort != 'A' || !sInputBuf.size() ) { return 0; }
	sInputBuf.lock();
	/////////////////
	byte ch = sInputBuf.c_str()[0];
	std::string sTmp = &sInputBuf.c_str()[1];
	sInputBuf = sTmp;
	/////////////////
	sInputBuf.unlock();
	return ch;
};

///////////////////////////////////////////////////////////////////////////////

const char* const stub_Serial_Class::CLASS_NAME  = stub_Serial_Class_CLASS_NAME;
const char* const stub_Serial_Class::OBJECT_TYPE = stub_Serial_Class_OBJECT_TYPE;

unsigned int stub_Serial_Class::nObjectsCount = 0;

InternalsPointersVectorType  stub_Serial_Class::vInternalsPointers;
InternalsPointersMapType     stub_Serial_Class::mInternalsPointers;

InternalsType  stub_Serial_Class::InternalsRoot = { OBJECT_TYPE, "Object Internals Root", "InternalsPointersVectorType*", "0x%X", &mInternalsPointers, (qword)&vInternalsPointers };

InternalsType  stub_Serial_ObjectConstructor = { "ObjectConstructor", NULL, "ObjectConstructorType*", "0x%X", NULL, (qword)stub_Serial_Class::ObjectConstructor };
AddInternal    i_stub_Serial_ObjectConstructor( &stub_Serial_Class::vInternalsPointers, &stub_Serial_ObjectConstructor );

ObjectClass* stub_Serial_Class::ObjectConstructor( ConfigInterface *pConfigI, const char *acAlias )
{
	return new stub_Serial_Class( pConfigI, acAlias );
}

///////////////////////////

stub_Serial_Class::stub_Serial_Class( ConfigInterface *pConfigI, const char *acAlias ):
ObjectClass( pConfigI, acDevice_SuperType, OBJECT_TYPE, acAlias, &InternalsRoot ),
AbstractDeviceClass( stub_architecture ),
stub_DeviceInterface( (ObjectClass*)this ),
ByteOutputSocket( acStream_SuperType, acByteStreamType, "", this ),
ByteInputSocket(  acStream_SuperType, acByteStreamType, "", this, fastdelegate::MakeDelegate(this, &stub_Serial_Class::ByteReceiver) )
{
	if( !nCreationSuccess ) { return; }

	nObjectsCount++;
	
	vPorts.push_back('C');
	vPorts.push_back('D');

SysOut("stub_Serial -- created !!!\n");	
}

///////////////////////////////////////////////////////////////////////////////

const char* const stub_Video_Class::CLASS_NAME  = stub_Video_Class_CLASS_NAME;
const char* const stub_Video_Class::OBJECT_TYPE = stub_Video_Class_OBJECT_TYPE;

unsigned int stub_Video_Class::nObjectsCount = 0;

InternalsPointersVectorType  stub_Video_Class::vInternalsPointers;
InternalsPointersMapType     stub_Video_Class::mInternalsPointers;

InternalsType  stub_Video_Class::InternalsRoot = { OBJECT_TYPE, "Object Internals Root", "InternalsPointersVectorType*", "0x%X", &mInternalsPointers, (qword)&vInternalsPointers };

InternalsType  stub_Video_ObjectConstructor = { "ObjectConstructor", NULL, "ObjectConstructorType*", "0x%X", NULL, (qword)stub_Video_Class::ObjectConstructor };
AddInternal    i_stub_Video_ObjectConstructor( &stub_Video_Class::vInternalsPointers, &stub_Video_ObjectConstructor );

ObjectClass* stub_Video_Class::ObjectConstructor( ConfigInterface *pConfigI, const char *acAlias )
{
	return new stub_Video_Class( pConfigI, acAlias );
}

///////////////////////////

stub_Video_Class::stub_Video_Class( ConfigInterface *pConfigI, const char *acAlias ):
ObjectClass( pConfigI, acDevice_SuperType, OBJECT_TYPE, acAlias, &InternalsRoot ),
FrameMonitorClass( pConfigI, acAlias ),
AbstractDeviceClass( stub_architecture ),
stub_DeviceInterface( (ObjectClass*)this ),
Frame( 800, 550 )
{
	if( !nCreationSuccess ) { return; }

	nObjectsCount++;
	
	vPorts.push_back('B');    //   'B'

	ShowFrame( &Frame );

SysOut("stub_Video -- created !!!\n");	
SysOut("\n");
}


void stub_Video_Class::WF_ActiveStateChanged( bool bActive ) 
{
	if( bActive ) 
	{
		pActiveBus = (dword)pBus;

		pConfigI->Disable_diKey_PassThru( DIK_NUMLOCK );
		pConfigI->Disable_diKey_PassThru( DIK_LMENU );
		pConfigI->Disable_diKey_PassThru( DIK_RMENU );
		pConfigI->Disable_diKey_PassThru( DIK_LWIN );
		pConfigI->Disable_diKey_PassThru( DIK_RWIN );
		pConfigI->Disable_diKey_PassThru( DIK_APPS );
		pConfigI->Disable_diKey_PassThru( DIK_F10 );
	}
	else          
	{
		InterlockedCompareExchange( &pActiveBus, NULL, (dword)pBus );
		
		pConfigI->Enable_diKey_PassThru( DIK_NUMLOCK );
		pConfigI->Enable_diKey_PassThru( DIK_LMENU );
		pConfigI->Enable_diKey_PassThru( DIK_RMENU );
		pConfigI->Enable_diKey_PassThru( DIK_LWIN );
		pConfigI->Enable_diKey_PassThru( DIK_RWIN );
		pConfigI->Enable_diKey_PassThru( DIK_APPS );
		pConfigI->Enable_diKey_PassThru( DIK_F10 );
	}
}

void stub_Video_Class::WritePort( byte uPort, byte uByte )
{
	if( uPort != 'B' ) { return; }
	
	Frame.dwFrameNum++;
	
	//          B,G,R
	memset( Frame.pBGR_Buf, uByte, Frame.nWidth*Frame.nHeight*4 );

}

void stub_Video_Class::WF_Tick()
{
	ShowFrame( &Frame );	
}